import factory

from oscar.core.loading import get_model

ConditionalOffer = get_model("offer", "ConditionalOffer")

__all__ = [
    "RangeFactory",
    "ConditionFactory",
    "BenefitFactory",
    "ConditionalOfferFactory",
]


class RangeFactory(factory.django.DjangoModelFactory):
    name = factory.Sequence(lambda n: "Range %d" % n)
    slug = factory.Sequence(lambda n: "range-%d" % n)

    class Meta:
        model = get_model("offer", "Range")

    @factory.post_generation
    def products(self, create, extracted, **kwargs):
        if not create or not extracted:
            return

        RangeProduct = get_model("offer", "RangeProduct")

        for product in extracted:
            RangeProduct.objects.create(product=product, range=self)


class BenefitFactory(factory.django.DjangoModelFactory):
    type = get_model("offer", "Benefit").PERCENTAGE
    value = 10
    max_affected_items = None
    range = factory.SubFactory(RangeFactory)

    class Meta:
        model = get_model("offer", "Benefit")


class ConditionFactory(factory.django.DjangoModelFactory):
    type = get_model("offer", "Condition").COUNT
    value = 10
    range = factory.SubFactory(RangeFactory)

    class Meta:
        model = get_model("offer", "Condition")


class ConditionalOfferFactory(factory.django.DjangoModelFactory):
    name = factory.Sequence(lambda n: "Test offer %d" % n)
    offer_type = ConditionalOffer.SITE
    benefit = factory.SubFactory(BenefitFactory)
    condition = factory.SubFactory(ConditionFactory)

    class Meta:
        model = get_model("offer", "ConditionalOffer")



# pylint: disable=attribute-defined-outside-init
from django import http
from django.conf import settings
from django.contrib import messages
from django.contrib.auth import login as auth_login
from django.contrib.auth import logout as auth_logout
from django.contrib.auth import update_session_auth_hash
from django.contrib.auth.forms import PasswordChangeForm
from django.core.exceptions import ObjectDoesNotExist
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse, reverse_lazy
from django.utils.translation import gettext_lazy as _
from django.views import generic

from oscar.apps.customer.utils import get_password_reset_url
from oscar.core.compat import get_user_model
from oscar.core.loading import get_class, get_classes, get_model, get_profile_class
from oscar.core.utils import safe_referrer
from oscar.views.generic import PostActionMixin

from . import signals
from .forms import AnonymousOrderForm
from .mixins import AnonymousOrderMixin

PageTitleMixin, RegisterUserMixin = get_classes(
    "customer.mixins", ["PageTitleMixin", "RegisterUserMixin"]
)
CustomerDispatcher = get_class("customer.utils", "CustomerDispatcher")
EmailAuthenticationForm, EmailUserCreationForm, OrderSearchForm = get_classes(
    "customer.forms",
    ["EmailAuthenticationForm", "EmailUserCreationForm", "OrderSearchForm"],
)
ProfileForm, ConfirmPasswordForm = get_classes(
    "customer.forms", ["ProfileForm", "ConfirmPasswordForm"]
)
UserAddressForm = get_class("address.forms", "UserAddressForm")
Order = get_model("order", "Order")
UserAddress = get_model("address", "UserAddress")
Email = get_model("communication", "Email")

User = get_user_model()


# =======
# Account
# =======


class AccountSummaryView(generic.RedirectView):
    """
    View that exists for legacy reasons and customisability. It commonly gets
    called when the user clicks on "Account" in the navbar.

    Oscar defaults to just redirecting to the profile summary page (and
    that redirect can be configured via OSCAR_ACCOUNT_REDIRECT_URL), but
    it's also likely you want to display an 'account overview' page or
    such like. The presence of this view allows just that, without
    having to change a lot of templates.
    """

    pattern_name = settings.OSCAR_ACCOUNTS_REDIRECT_URL
    permanent = False


class AccountRegistrationView(RegisterUserMixin, generic.FormView):
    form_class = EmailUserCreationForm
    template_name = "oscar/customer/registration.html"
    redirect_field_name = "next"

    def get(self, request, *args, **kwargs):
        if request.user.is_authenticated:
            return redirect(settings.LOGIN_REDIRECT_URL)
        return super().get(request, *args, **kwargs)

    def get_logged_in_redirect(self):
        return reverse("customer:summary")

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["initial"] = {
            "email": self.request.GET.get("email", ""),
            "redirect_url": self.request.GET.get(self.redirect_field_name, ""),
        }
        kwargs["host"] = self.request.get_host()
        return kwargs

    def get_context_data(self, *args, **kwargs):
        ctx = super().get_context_data(*args, **kwargs)
        ctx["cancel_url"] = safe_referrer(self.request, "")
        return ctx

    def form_valid(self, form):
        self.register_user(form)
        return redirect(form.cleaned_data["redirect_url"])


class AccountAuthView(RegisterUserMixin, generic.TemplateView):
    """
    This is actually a slightly odd double form view that allows a customer to
    either login or register.
    """

    template_name = "oscar/customer/login_registration.html"
    login_prefix, registration_prefix = "login", "registration"
    login_form_class = EmailAuthenticationForm
    registration_form_class = EmailUserCreationForm
    redirect_field_name = "next"

    def get(self, request, *args, **kwargs):
        if request.user.is_authenticated:
            return redirect(settings.LOGIN_REDIRECT_URL)
        return super().get(request, *args, **kwargs)

    def get_context_data(self, *args, **kwargs):
        ctx = super().get_context_data(*args, **kwargs)
        if "login_form" not in kwargs:
            ctx["login_form"] = self.get_login_form()
        if "registration_form" not in kwargs:
            ctx["registration_form"] = self.get_registration_form()
        return ctx

    def post(self, request, *args, **kwargs):
        # Use the name of the submit button to determine which form to validate
        if "login_submit" in request.POST:
            return self.validate_login_form()
        elif "registration_submit" in request.POST:
            return self.validate_registration_form()
        return http.HttpResponseBadRequest()

    # LOGIN

    def get_login_form(self, bind_data=False):
        return self.login_form_class(**self.get_login_form_kwargs(bind_data))

    def get_login_form_kwargs(self, bind_data=False):
        kwargs = {}
        kwargs["request"] = self.request
        kwargs["host"] = self.request.get_host()
        kwargs["prefix"] = self.login_prefix
        kwargs["initial"] = {
            "redirect_url": self.request.GET.get(self.redirect_field_name, ""),
        }
        if bind_data and self.request.method in ("POST", "PUT"):
            kwargs.update(
                {
                    "data": self.request.POST,
                    "files": self.request.FILES,
                }
            )
        return kwargs

    def validate_login_form(self):
        form = self.get_login_form(bind_data=True)
        if form.is_valid():
            user = form.get_user()

            # Grab a reference to the session ID before logging in
            old_session_key = self.request.session.session_key

            auth_login(self.request, form.get_user())

            # Raise signal robustly (we don't want exceptions to crash the
            # request handling). We use a custom signal as we want to track the
            # session key before calling login (which cycles the session ID).
            signals.user_logged_in.send_robust(
                sender=self,
                request=self.request,
                user=user,
                old_session_key=old_session_key,
            )

            msg = self.get_login_success_message(form)
            if msg:
                messages.success(self.request, msg)

            return redirect(self.get_login_success_url(form))

        ctx = self.get_context_data(login_form=form)
        return self.render_to_response(ctx)

    # pylint: disable=unused-argument
    def get_login_success_message(self, form):
        return _("Welcome back")

    def get_login_success_url(self, form):
        redirect_url = form.cleaned_data["redirect_url"]
        if redirect_url:
            return redirect_url

        # Redirect staff members to dashboard as that's the most likely place
        # they'll want to visit if they're logging in.
        if self.request.user.is_staff:
            return reverse("dashboard:index")

        return settings.LOGIN_REDIRECT_URL

    # REGISTRATION

    def get_registration_form(self, bind_data=False):
        return self.registration_form_class(
            **self.get_registration_form_kwargs(bind_data)
        )

    def get_registration_form_kwargs(self, bind_data=False):
        kwargs = {}
        kwargs["host"] = self.request.get_host()
        kwargs["prefix"] = self.registration_prefix
        kwargs["initial"] = {
            "redirect_url": self.request.GET.get(self.redirect_field_name, ""),
        }
        if bind_data and self.request.method in ("POST", "PUT"):
            kwargs.update(
                {
                    "data": self.request.POST,
                    "files": self.request.FILES,
                }
            )
        return kwargs

    def validate_registration_form(self):
        form = self.get_registration_form(bind_data=True)
        if form.is_valid():
            self.register_user(form)

            msg = self.get_registration_success_message(form)
            messages.success(self.request, msg)

            return redirect(self.get_registration_success_url(form))

        ctx = self.get_context_data(registration_form=form)
        return self.render_to_response(ctx)

    # pylint: disable=unused-argument
    def get_registration_success_message(self, form):
        return _("Thanks for registering!")

    def get_registration_success_url(self, form):
        redirect_url = form.cleaned_data["redirect_url"]
        if redirect_url:
            return redirect_url

        return settings.LOGIN_REDIRECT_URL


class LogoutView(generic.RedirectView):
    url = settings.OSCAR_HOMEPAGE
    permanent = False

    def get(self, request, *args, **kwargs):
        auth_logout(request)
        response = super().get(request, *args, **kwargs)

        for cookie in settings.OSCAR_COOKIES_DELETE_ON_LOGOUT:
            response.delete_cookie(cookie)

        return response


# =============
# Profile
# =============


class ProfileView(PageTitleMixin, generic.TemplateView):
    template_name = "oscar/customer/profile/profile.html"
    page_title = _("Profile")
    active_tab = "profile"

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["profile_fields"] = self.get_profile_fields(self.request.user)
        return ctx

    def get_profile_fields(self, user):
        field_data = []

        # Check for custom user model
        for field_name in User._meta.additional_fields:
            field_data.append(self.get_model_field_data(user, field_name))

        # Check for profile class
        profile_class = get_profile_class()
        if profile_class:
            try:
                profile = profile_class.objects.get(user=user)
            except ObjectDoesNotExist:
                profile = profile_class(user=user)

            field_names = [f.name for f in profile._meta.local_fields]
            for field_name in field_names:
                if field_name in ("user", "id"):
                    continue
                field_data.append(self.get_model_field_data(profile, field_name))

        return field_data

    def get_model_field_data(self, model_class, field_name):
        """
        Extract the verbose name and value for a model's field value
        """
        field = model_class._meta.get_field(field_name)
        if field.choices:
            value = getattr(model_class, "get_%s_display" % field_name)()
        else:
            value = getattr(model_class, field_name)
        return {
            "name": getattr(field, "verbose_name"),
            "value": value,
        }


class ProfileUpdateView(PageTitleMixin, generic.FormView):
    form_class = ProfileForm
    template_name = "oscar/customer/profile/profile_form.html"
    page_title = _("Edit Profile")
    active_tab = "profile"
    success_url = reverse_lazy("customer:profile-view")

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["user"] = self.request.user
        return kwargs

    def form_valid(self, form):
        # Grab current user instance before we save form.  We may need this to
        # send a warning email if the email address is changed.
        try:
            old_user = User.objects.get(id=self.request.user.id)
        except User.DoesNotExist:
            old_user = None

        form.save()

        # We have to look up the email address from the form's
        # cleaned data because the object created by form.save() can
        # either be a user or profile instance depending whether a profile
        # class has been specified by the AUTH_PROFILE_MODULE setting.
        new_email = form.cleaned_data.get("email")
        if new_email and old_user and new_email != old_user.email:
            # Email address has changed - send a confirmation email to the old
            # address including a password reset link in case this is a
            # suspicious change.
            self.send_email_changed_email(old_user, new_email)

        messages.success(self.request, _("Profile updated"))
        return redirect(self.get_success_url())

    def send_email_changed_email(self, old_user, new_email):
        user = self.request.user
        extra_context = {
            "user": user,
            "reset_url": get_password_reset_url(user),
            "new_email": new_email,
            "request": self.request,
        }
        CustomerDispatcher().send_email_changed_email_for_user(old_user, extra_context)


class ProfileDeleteView(PageTitleMixin, generic.FormView):
    form_class = ConfirmPasswordForm
    template_name = "oscar/customer/profile/profile_delete.html"
    page_title = _("Delete profile")
    active_tab = "profile"
    success_url = settings.OSCAR_HOMEPAGE

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["user"] = self.request.user
        return kwargs

    def form_valid(self, form):
        self.request.user.delete()
        messages.success(
            self.request,
            _("Your profile has now been deleted. Thanks for using the site."),
        )
        return redirect(self.get_success_url())


class ChangePasswordView(PageTitleMixin, generic.FormView):
    form_class = PasswordChangeForm
    template_name = "oscar/customer/profile/change_password_form.html"
    page_title = _("Change Password")
    active_tab = "profile"
    success_url = reverse_lazy("customer:profile-view")

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["user"] = self.request.user
        return kwargs

    def form_valid(self, form):
        form.save()
        update_session_auth_hash(self.request, self.request.user)
        messages.success(self.request, _("Password updated"))

        self.send_password_changed_email()

        return redirect(self.get_success_url())

    def send_password_changed_email(self):
        user = self.request.user
        extra_context = {
            "user": user,
            "reset_url": get_password_reset_url(self.request.user),
            "request": self.request,
        }
        CustomerDispatcher().send_password_changed_email_for_user(user, extra_context)


# =============
# Email history
# =============


class EmailHistoryView(PageTitleMixin, generic.ListView):
    context_object_name = "emails"
    template_name = "oscar/communication/email/email_list.html"
    paginate_by = settings.OSCAR_EMAILS_PER_PAGE
    page_title = _("Email History")
    active_tab = "emails"

    def get_queryset(self):
        """
        Return Queryset of :py:class:`Email <oscar.apps.customer.abstract_models.AbstractEmail>`
        instances, that has been sent to the currently authenticated user.
        """
        return Email._default_manager.filter(user=self.request.user)


class EmailDetailView(PageTitleMixin, generic.DetailView):
    """Customer email"""

    template_name = "oscar/communication/email/email_detail.html"
    context_object_name = "email"
    active_tab = "emails"

    def get_object(self, queryset=None):
        return get_object_or_404(
            Email, user=self.request.user, id=self.kwargs["email_id"]
        )

    def get_page_title(self):
        """Append email subject to page title"""
        return "%s: %s" % (_("Email"), self.object.subject)


# =============
# Order history
# =============


class OrderHistoryView(PageTitleMixin, generic.ListView):
    """
    Customer order history
    """

    context_object_name = "orders"
    template_name = "oscar/customer/order/order_list.html"
    paginate_by = settings.OSCAR_ORDERS_PER_PAGE
    model = Order
    form_class = OrderSearchForm
    page_title = _("Order History")
    active_tab = "orders"

    def get(self, request, *args, **kwargs):
        if "date_from" in request.GET:
            self.form = self.form_class(self.request.GET)
            if not self.form.is_valid():
                self.object_list = self.get_queryset()
                ctx = self.get_context_data(object_list=self.object_list)
                return self.render_to_response(ctx)
            data = self.form.cleaned_data

            # If the user has just entered an order number, try and look it up
            # and redirect immediately to the order detail page.
            if data["order_number"] and not (data["date_to"] or data["date_from"]):
                try:
                    order = Order.objects.get(
                        number=data["order_number"], user=self.request.user
                    )
                except Order.DoesNotExist:
                    pass
                else:
                    return redirect("customer:order", order_number=order.number)
        else:
            self.form = self.form_class()
        return super().get(request, *args, **kwargs)

    def get_queryset(self):
        """
        Return Queryset of :py:class:`Order <oscar.apps.order.abstract_models.AbstractOrder>`
        instances for the currently authenticated user.
        """
        qs = self.model._default_manager.filter(user=self.request.user)
        if self.form.is_bound and self.form.is_valid():
            qs = qs.filter(**self.form.get_filters())
        return qs

    def get_context_data(self, *args, **kwargs):
        ctx = super().get_context_data(*args, **kwargs)
        ctx["form"] = self.form
        return ctx


class OrderDetailView(PageTitleMixin, PostActionMixin, generic.DetailView):
    model = Order
    active_tab = "orders"

    def get_template_names(self):
        return ["oscar/customer/order/order_detail.html"]

    def get_page_title(self):
        """
        Order number as page title
        """
        return "%s #%s" % (_("Order"), self.object.number)

    def get_object(self, queryset=None):
        return get_object_or_404(
            self.model, user=self.request.user, number=self.kwargs["order_number"]
        )

    def do_reorder(self, order):
        """
        'Re-order' a previous order.

        This puts the contents of the previous order into your basket
        """
        # Collect lines to be added to the basket and any warnings for lines
        # that are no longer available.
        basket = self.request.basket
        lines_to_add = []
        warnings = []
        for line in order.lines.all():
            is_available, reason = line.is_available_to_reorder(
                basket, self.request.strategy
            )
            if is_available:
                lines_to_add.append(line)
            else:
                warnings.append(reason)

        # Check whether the number of items in the basket won't exceed the
        # maximum.
        total_quantity = sum([line.quantity for line in lines_to_add])
        is_quantity_allowed, reason = basket.is_quantity_allowed(total_quantity)
        if not is_quantity_allowed:
            messages.warning(self.request, reason)
            self.response = redirect("customer:order-list")
            return

        # Add any warnings
        for warning in warnings:
            messages.warning(self.request, warning)

        for line in lines_to_add:
            options = []
            for attribute in line.attributes.all():
                if attribute.option:
                    options.append(
                        {"option": attribute.option, "value": attribute.value}
                    )
            basket.add_product(line.product, line.quantity, options)

        if len(lines_to_add) > 0:
            self.response = redirect("basket:summary")
            messages.info(
                self.request,
                _(
                    "All available lines from order %(number)s "
                    "have been added to your basket"
                )
                % {"number": order.number},
            )
        else:
            self.response = redirect("customer:order-list")
            messages.warning(
                self.request,
                _(
                    "It is not possible to re-order order %(number)s "
                    "as none of its lines are available to purchase"
                )
                % {"number": order.number},
            )


class OrderLineView(PostActionMixin, generic.DetailView):
    """Customer order line"""

    def get_object(self, queryset=None):
        order = get_object_or_404(
            Order, user=self.request.user, number=self.kwargs["order_number"]
        )
        return order.lines.get(id=self.kwargs["line_id"])

    def do_reorder(self, line):
        self.response = redirect("customer:order", self.kwargs["order_number"])
        basket = self.request.basket

        line_available_to_reorder, reason = line.is_available_to_reorder(
            basket, self.request.strategy
        )

        if not line_available_to_reorder:
            messages.warning(self.request, reason)
            return

        # We need to pass response to the get_or_create... method
        # as a new basket might need to be created
        self.response = redirect("basket:summary")

        # Convert line attributes into basket options
        options = []
        for attribute in line.attributes.all():
            if attribute.option:
                options.append({"option": attribute.option, "value": attribute.value})
        basket.add_product(line.product, line.quantity, options)

        if line.quantity > 1:
            msg = _(
                "%(qty)d copies of '%(product)s' have been added to your basket"
            ) % {"qty": line.quantity, "product": line.product}
        else:
            msg = _("'%s' has been added to your basket") % line.product

        messages.info(self.request, msg)


class AnonymousOrderFormView(AnonymousOrderMixin, generic.FormView):
    form_class = AnonymousOrderForm
    template_name = "oscar/customer/anon_order_form.html"

    def form_valid(self, form):
        self.request.session["anon_order_email"] = form.cleaned_data["email"]
        if not self.get_object():
            form.add_error(None, _("Order not found"))
            return self.form_invalid(form)
        return super().form_valid(form)

    def get_success_url(self):
        return reverse(
            "customer:anon-order",
            kwargs={"order_number": self.order_number, "hash": self.hash},
        )


class AnonymousOrderDetailView(AnonymousOrderMixin, generic.DetailView):
    model = Order
    template_name = "oscar/customer/anon_order.html"

    def dispatch(self, request, *args, **kwargs):
        if not self.get_object():
            return redirect(
                "customer:anon-order-form",
                order_number=self.order_number,
                hash=self.hash,
            )
        return super().dispatch(request, *args, **kwargs)


# ------------
# Address book
# ------------


class AddressListView(PageTitleMixin, generic.ListView):
    """Customer address book"""

    context_object_name = "addresses"
    template_name = "oscar/customer/address/address_list.html"
    paginate_by = settings.OSCAR_ADDRESSES_PER_PAGE
    active_tab = "addresses"
    page_title = _("Address Book")

    def get_queryset(self):
        """Return customer's addresses"""
        return UserAddress._default_manager.filter(user=self.request.user)


class AddressCreateView(PageTitleMixin, generic.CreateView):
    form_class = UserAddressForm
    model = UserAddress
    template_name = "oscar/customer/address/address_form.html"
    active_tab = "addresses"
    page_title = _("Add a new address")
    success_url = reverse_lazy("customer:address-list")

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["user"] = self.request.user
        return kwargs

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["title"] = _("Add a new address")
        return ctx

    def get_success_url(self):
        messages.success(self.request, _("Address '%s' created") % self.object.summary)
        return super().get_success_url()


class AddressUpdateView(PageTitleMixin, generic.UpdateView):
    form_class = UserAddressForm
    model = UserAddress
    template_name = "oscar/customer/address/address_form.html"
    active_tab = "addresses"
    page_title = _("Edit address")
    success_url = reverse_lazy("customer:address-list")

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["user"] = self.request.user
        return kwargs

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["title"] = _("Edit address")
        return ctx

    def get_queryset(self):
        return self.request.user.addresses.all()

    def get_success_url(self):
        messages.success(self.request, _("Address '%s' updated") % self.object.summary)
        return super().get_success_url()


class AddressDeleteView(PageTitleMixin, generic.DeleteView):
    model = UserAddress
    template_name = "oscar/customer/address/address_delete.html"
    page_title = _("Delete address?")
    active_tab = "addresses"
    context_object_name = "address"
    success_url = reverse_lazy("customer:address-list")

    def get_queryset(self):
        return UserAddress._default_manager.filter(user=self.request.user)

    def get_success_url(self):
        messages.success(self.request, _("Address '%s' deleted") % self.object.summary)
        return super().get_success_url()


class AddressChangeStatusView(generic.RedirectView):
    """
    Sets an address as default_for_(billing|shipping)
    """

    url = reverse_lazy("customer:address-list")
    permanent = False

    def get(self, request, *args, pk=None, action=None, **kwargs):
        address = get_object_or_404(UserAddress, user=self.request.user, pk=pk)
        #  We don't want the user to set an address as the default shipping
        #  address, though they should be able to set it as their billing
        #  address.
        if address.country.is_shipping_country:
            setattr(address, "is_%s" % action, True)
        elif action == "default_for_billing":
            setattr(address, "is_default_for_billing", True)
        else:
            messages.error(request, _("We do not ship to this country"))
        address.save()
        return super().get(request, *args, **kwargs)



from django.conf import settings
from oscar.core.loading import get_class
from haystack.query import SearchQuerySet
from haystack.exceptions import MissingDependency

try:
    from haystack.backends.whoosh_backend import WhooshSearchQuery
except MissingDependency:
    WhooshSearchQuery = None

from purl import URL

is_solr_supported = get_class("search.features", "is_solr_supported")


def base_sqs():
    """
    Return the base SearchQuerySet for Haystack searches.
    """
    sqs = SearchQuerySet()
    for facet in settings.OSCAR_SEARCH_FACETS["fields"].values():
        options = facet.get("options", {})
        sqs = sqs.facet(facet["field"], **options)
    for facet in settings.OSCAR_SEARCH_FACETS["queries"].values():
        for query in facet["queries"]:
            sqs = sqs.query_facet(facet["field"], query[1])

    sqs = sqs.filter_and(is_public="true", structure__in=["standalone", "parent"])
    return sqs


class FacetMunger(object):
    def __init__(self, path, selected_multi_facets, facet_counts, query_type=None):
        self.base_url = URL(path)
        self.selected_facets = selected_multi_facets
        self.facet_counts = facet_counts
        self.query_type = query_type

    def facet_data(self):
        facet_data = {}
        # Haystack can return an empty dict for facet_counts when e.g. Solr
        # isn't running. Skip facet munging in that case.
        if self.facet_counts:
            self.munge_field_facets(facet_data)
            if self.query_type is not None and self.query_type not in [
                WhooshSearchQuery
            ]:
                self.munge_query_facets(facet_data)
        return facet_data

    def munge_field_facets(self, clean_data):
        for key, facet in settings.OSCAR_SEARCH_FACETS["fields"].items():
            self.munge_field_facet(key, facet, clean_data)

    def munge_field_facet(self, key, facet, clean_data):
        clean_data[key] = {"name": facet["name"], "results": []}
        for field_value, count in self.facet_counts["fields"][key]:
            field_name = facet["field"]
            is_faceted_already = field_name in self.selected_facets
            datum = {
                "name": field_value,
                "count": count,
                # We don't show facet counts if a this field is already being
                # faceted (as we don't know them)
                "show_count": not is_faceted_already,
                "disabled": count == 0 and not is_faceted_already,
                "selected": False,
            }
            if field_value in self.selected_facets.get(field_name, []):
                # This filter is selected - build the 'deselect' URL
                datum["selected"] = True
                url = self.base_url.remove_query_param(
                    "selected_facets", "%s:%s" % (field_name, field_value)
                )
                datum["deselect_url"] = self.strip_pagination(url)
            else:
                # This filter is not selected - built the 'select' URL
                url = self.base_url.append_query_param(
                    "selected_facets", "%s:%s" % (field_name, field_value)
                )
                datum["select_url"] = self.strip_pagination(url)

            clean_data[key]["results"].append(datum)

    def munge_query_facets(self, clean_data):
        for key, facet in settings.OSCAR_SEARCH_FACETS["queries"].items():
            self.munge_query_facet(key, facet, clean_data)

    def munge_query_facet(self, key, facet, clean_data):
        clean_data[key] = {"name": facet["name"], "results": []}
        # Loop over the queries in OSCAR_SEARCH_FACETS rather than the returned
        # facet information from the search backend.
        for field_value, query in facet["queries"]:
            field_name = facet["field"]
            is_faceted_already = field_name in self.selected_facets
            match = "%s:%s" % (field_name, query)
            # Solr expects "_exact" in field names
            match_exact = (
                f"{field_name}_exact:{query}" if is_solr_supported() else match
            )
            if match_exact not in self.facet_counts["queries"]:
                # This query was not returned
                datum = {
                    "name": field_value,
                    "count": 0,
                    "show_count": True,
                    "disabled": True,
                }
            else:
                count = self.facet_counts["queries"][match_exact]
                datum = {
                    "name": field_value,
                    "count": count,
                    "show_count": not is_faceted_already,
                    "disabled": count == 0 and not is_faceted_already,
                    "selected": False,
                }
                if query in self.selected_facets.get(field_name, []):
                    # Selected
                    datum["selected"] = True
                    datum["show_count"] = True
                    url = self.base_url.remove_query_param("selected_facets", match)
                    datum["deselect_url"] = self.strip_pagination(url)
                else:
                    url = self.base_url.append_query_param("selected_facets", match)
                    datum["select_url"] = self.strip_pagination(url)
            clean_data[key]["results"].append(datum)

    def strip_pagination(self, url):
        if url.has_query_param("page"):
            url = url.remove_query_param("page")
        return url.as_string()



from django.urls import path
from django.utils.translation import gettext_lazy as _

from oscar.core.application import OscarDashboardConfig
from oscar.core.loading import get_class


class PagesDashboardConfig(OscarDashboardConfig):
    label = "pages_dashboard"
    name = "oscar.apps.dashboard.pages"
    verbose_name = _("Pages dashboard")

    default_permissions = [
        "is_staff",
    ]

    def configure_permissions(self):
        DashboardPermission = get_class("dashboard.permissions", "DashboardPermission")

        self.permissions_map = {
            "page-list": DashboardPermission.get("flat_page"),
            "page-create": DashboardPermission.get("flat_page"),
            "page-update": DashboardPermission.get("flat_page"),
            "page-delete": DashboardPermission.get("flat_page"),
        }

    # pylint: disable=attribute-defined-outside-init
    def ready(self):
        self.list_view = get_class("dashboard.pages.views", "PageListView")
        self.create_view = get_class("dashboard.pages.views", "PageCreateView")
        self.update_view = get_class("dashboard.pages.views", "PageUpdateView")
        self.delete_view = get_class("dashboard.pages.views", "PageDeleteView")
        self.configure_permissions()

    def get_urls(self):
        """
        Get URL patterns defined for flatpage management application.
        """
        urls = [
            path("", self.list_view.as_view(), name="page-list"),
            path("create/", self.create_view.as_view(), name="page-create"),
            path("update/<str:pk>/", self.update_view.as_view(), name="page-update"),
            path("delete/<str:pk>/", self.delete_view.as_view(), name="page-delete"),
        ]
        return self.post_process_urls(urls)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        ('partner', '0001_initial'),
        ('catalogue', '0001_initial'),
        ('basket', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AddField(
            model_name='lineattribute',
            name='option',
            field=models.ForeignKey(verbose_name='Option', to='catalogue.Option', on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='line',
            name='basket',
            field=models.ForeignKey(verbose_name='Basket', related_name='lines', to='basket.Basket', on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='line',
            name='product',
            field=models.ForeignKey(verbose_name='Product', related_name='basket_lines', to='catalogue.Product', on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='line',
            name='stockrecord',
            field=models.ForeignKey(related_name='basket_lines', to='partner.StockRecord', on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AlterUniqueTogether(
            name='line',
            unique_together=set([('basket', 'line_reference')]),
        ),
        migrations.AddField(
            model_name='basket',
            name='owner',
            field=models.ForeignKey(verbose_name='Owner', related_name='baskets', to=settings.AUTH_USER_MODEL, null=True, on_delete=models.CASCADE),
            preserve_default=True,
        ),
    ]



# Generated by Django 3.2.9 on 2022-01-25 19:01
import logging

from django.db import migrations
from django.db.models import CharField, Count, Value
from django.db.models.functions import Concat

logger = logging.getLogger(__name__)


def get_value(attribute_value):
    typ = attribute_value.attribute.type
    field_name = "value_%s" % typ
    value = getattr(attribute_value, field_name)
    if hasattr(value, "all"):
        value = value.all()
    return value

def remove_duplicate_attributes(apps, schema_editor):
    """
    Removes duplicate attributes that have the same code and product class.
    """
    ProductAttribute = apps.get_model('catalogue', 'ProductAttribute')
    ProductAttributeValue = apps.get_model('catalogue', 'ProductAttributeValue')
    ProductClass = apps.get_model("catalogue", "ProductClass")

    # Instead of iterating over all attributes, we concat the code and product class pk
    # with a "|" so we can find duplicate attributes in one query.
    duplicate_attributes = ProductAttribute.objects.filter(product_class__isnull=False).annotate(
        code_and_product_class=Concat('code', Value('|'), 'product_class__pk', output_field=CharField())
    ).values('code_and_product_class').annotate(
        same_code_count=Count('code_and_product_class')
    ).filter(same_code_count__gt=1)

    for attribute in duplicate_attributes:
        attribute_code, product_class_pk = attribute["code_and_product_class"].split("|")
        product_class = ProductClass.objects.get(pk=product_class_pk)
        attributes = ProductAttribute.objects.filter(
            code=attribute_code,
            product_class=product_class
        )
        used_attributes = attributes.filter(productattributevalue__isnull=False)
        used_attribute_count = used_attributes.distinct().count()

        # In most cases, the used attributes count will be one or zero as
        # the dashboard will always show one attribute. If the used attribute
        # count is one, we exclude this from attributes and remove the others.
        # If it's zero, we pick the last created and delete others.
        if used_attribute_count == 1:
            attributes.exclude(pk=used_attributes.first().pk).delete()
            continue
        elif used_attribute_count == 0:
            attributes.exclude(pk=attributes.last().pk).delete()
            continue

        # If we found multiple attributes that have values linked to them,
        # we must move them to one attribute and then delete the others.
        # We can only do this if the value_types are all the same!
        ASSERTION_MESSAGE = """Duplicate attribute found with code: %s but different types!
        You could fix this by renaming the duplicate codes or by matching all types to one
        type and update the attribute values accordingly for their new type. After that you can
        re-run the migration.""" % attribute_code
        assert used_attributes.values("type").distinct().count() == 1, ASSERTION_MESSAGE

        # Choose one attribute that will be used to move to and others to be deleted.
        to_be_used_attribute = used_attributes.first()
        to_be_deleted_attributes = used_attributes.exclude(pk=to_be_used_attribute.pk)

        for attribute in to_be_deleted_attributes:
            for attribute_value in attribute.productattributevalue_set.all():
                product = attribute_value.product

                # ProductAttributeValue has a unique together constraint on 'product' and 'attribute'.
                # This means, if the product of the current 'attribute_value' already has a ProductAttributeValue
                # linked to the 'to_be_used_attribute' attribute, we can't update the attribute on the
                # 'attribute_value' as this would raise an IntegrityError.
                to_be_used_attribute_value = to_be_used_attribute.productattributevalue_set.filter(product=product).first()
                if not to_be_used_attribute_value:
                    attribute_value.attribute = to_be_used_attribute
                    attribute_value.save()
                else:
                    msg = """Product with ID '%s' had more than one attribute value linked to an attribute
                    with code '%s'. We've kept the value '%s' and removed the value '%s' as this is the one you
                    would see in the dashboard when editing the product.
                    """ % (
                        product.id,
                        attribute.code,
                        get_value(to_be_used_attribute_value),
                        get_value(attribute_value)
                    )
                    logger.warning(msg)

            # Once the attribute values have been updated, we can safely remove the attribute instance.
            attribute.delete()



class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0023_auto_20210824_1414'),
    ]

    operations = [
        migrations.RunPython(remove_duplicate_attributes, migrations.RunPython.noop)
    ]



from importlib import import_module

from django.core import exceptions
from django.urls import reverse

# pylint: disable=W0611
from oscar.apps.offer.applicator import (
    Applicator,
)  # backwards-compat


def range_anchor(product_range):
    return '<a href="%s">%s</a>' % (
        reverse("dashboard:range-update", kwargs={"pk": product_range.pk}),
        product_range.name,
    )


# pylint: disable=unused-argument
def unit_price(offer, line):
    """
    Return the relevant price for a given basket line.

    This is required so offers can apply in circumstances where tax isn't known
    """
    return line.unit_effective_price


def load_proxy(proxy_class):
    module, classname = proxy_class.rsplit(".", 1)
    try:
        mod = import_module(module)
    except ImportError as e:
        raise exceptions.ImproperlyConfigured(
            "Error importing module %s: %s" % (module, e)
        )
    try:
        return getattr(mod, classname)
    except AttributeError:
        raise exceptions.ImproperlyConfigured(
            "Module %s does not define a %s" % (module, classname)
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.12 on 2018-04-13 08:11
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('voucher', '0005_auto_20180402_1425'),
    ]

    operations = [
        migrations.AlterField(
            model_name='voucherset',
            name='count',
            field=models.PositiveIntegerField(verbose_name='Number of vouchers'),
        ),
    ]



# Generated by Django 4.2.2 on 2023-07-19 13:16

from django.db import migrations, models
import django.db.models.deletion

from django.utils.module_loading import import_string
from django.conf import settings

models_AutoField = import_string(settings.DEFAULT_AUTO_FIELD)


class Migration(migrations.Migration):
    dependencies = [
        ("order", "0014_tax_code"),
    ]

    operations = [
        migrations.CreateModel(
            name="OrderLineDiscount",
            fields=[
                (
                    "id",
                    models_AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("is_incl_tax", models.BooleanField()),
                (
                    "amount",
                    models.DecimalField(
                        decimal_places=2,
                        default=0,
                        max_digits=12,
                        verbose_name="Line discount (excl. tax)",
                    ),
                ),
                (
                    "line",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="discounts",
                        to="order.line",
                        verbose_name="Line",
                    ),
                ),
                (
                    "order_discount",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="discount_lines",
                        to="order.orderdiscount",
                        verbose_name="Order discount",
                    ),
                ),
            ],
            options={
                "verbose_name": "Order line discount",
                "verbose_name_plural": "Order line discounts",
                "ordering": ["pk"],
                "abstract": False,
            },
        ),
    ]



# pylint: disable=W0401
# -*- coding: utf-8 -*-
from oscar.core.loading import is_model_registered

from .abstract_models import AbstractLine, AbstractWishListSharedEmail, AbstractWishList

__all__ = []


if not is_model_registered("wishlists", "WishList"):

    class WishList(AbstractWishList):
        pass

    __all__.append("WishList")


if not is_model_registered("wishlists", "Line"):

    class Line(AbstractLine):
        pass

    __all__.append("Line")


if not is_model_registered("wishlists", "WishListSharedEmail"):

    class WishListSharedEmail(AbstractWishListSharedEmail):
        pass

    __all__.append("WishListSharedEmail")



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('address', '0002_auto_20150927_1547'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='useraddress',
            options={'ordering': ['-num_orders_as_shipping_address'], 'verbose_name': 'User address',
                     'verbose_name_plural': 'User addresses'},
        ),
        migrations.AlterField(
            model_name='useraddress',
            name='num_orders_as_billing_address',
            field=models.PositiveIntegerField(default=0, verbose_name='Number of Orders as Shipping Address'),
        ),
        migrations.AlterField(
            model_name='useraddress',
            name='num_orders_as_shipping_address',
            field=models.PositiveIntegerField(default=0, verbose_name='Number of Orders as Billing Address'),
        ),
    ]



from django.conf import settings

from oscar.core.loading import get_model


def get_default_review_status():
    ProductReview = get_model("reviews", "ProductReview")

    if settings.OSCAR_MODERATE_REVIEWS:
        return ProductReview.FOR_MODERATION

    return ProductReview.APPROVED



from django.urls import path, re_path
from django.utils.translation import gettext_lazy as _

from oscar.core.application import OscarConfig
from oscar.core.loading import get_class


class OfferConfig(OscarConfig):
    label = "offer"
    name = "oscar.apps.offer"
    verbose_name = _("Offer")

    namespace = "offer"

    # pylint: disable=attribute-defined-outside-init, W0611, W0201
    def ready(self):
        from . import receivers

        self.detail_view = get_class("offer.views", "OfferDetailView")
        self.list_view = get_class("offer.views", "OfferListView")

    def get_urls(self):
        urls = [
            path("", self.list_view.as_view(), name="list"),
            re_path(r"^(?P<slug>[\w-]+)/$", self.detail_view.as_view(), name="detail"),
        ]
        return self.post_process_urls(urls)



from django.urls import path
from django.utils.translation import gettext_lazy as _

from oscar.core.application import OscarDashboardConfig
from oscar.core.loading import get_class


class VouchersDashboardConfig(OscarDashboardConfig):
    label = "vouchers_dashboard"
    name = "oscar.apps.dashboard.vouchers"
    verbose_name = _("Vouchers dashboard")

    default_permissions = [
        "is_staff",
    ]

    def configure_permissions(self):
        DashboardPermission = get_class("dashboard.permissions", "DashboardPermission")

        self.permissions_map = {
            "voucher-list": DashboardPermission.get("voucher"),
            "voucher-create": DashboardPermission.get("voucher"),
            "voucher-update": DashboardPermission.get("voucher"),
            "voucher-delete": DashboardPermission.get("voucher"),
            "voucher-stats": DashboardPermission.get("voucher"),
            "voucher-set-list": DashboardPermission.get("voucher"),
            "voucher-set-create": DashboardPermission.get("voucher"),
            "voucher-set-update": DashboardPermission.get("voucher"),
            "voucher-set-detail": DashboardPermission.get("voucher"),
            "voucher-set-download": DashboardPermission.get("voucher"),
            "voucher-set-delete": DashboardPermission.get("voucher"),
        }

    # pylint: disable=attribute-defined-outside-init
    def ready(self):
        self.list_view = get_class("dashboard.vouchers.views", "VoucherListView")
        self.create_view = get_class("dashboard.vouchers.views", "VoucherCreateView")
        self.update_view = get_class("dashboard.vouchers.views", "VoucherUpdateView")
        self.delete_view = get_class("dashboard.vouchers.views", "VoucherDeleteView")
        self.stats_view = get_class("dashboard.vouchers.views", "VoucherStatsView")

        self.set_list_view = get_class("dashboard.vouchers.views", "VoucherSetListView")
        self.set_create_view = get_class(
            "dashboard.vouchers.views", "VoucherSetCreateView"
        )
        self.set_update_view = get_class(
            "dashboard.vouchers.views", "VoucherSetUpdateView"
        )
        self.set_detail_view = get_class(
            "dashboard.vouchers.views", "VoucherSetDetailView"
        )
        self.set_download_view = get_class(
            "dashboard.vouchers.views", "VoucherSetDownloadView"
        )
        self.set_delete_view = get_class(
            "dashboard.vouchers.views", "VoucherSetDeleteView"
        )
        self.configure_permissions()

    def get_urls(self):
        urls = [
            path("", self.list_view.as_view(), name="voucher-list"),
            path("create/", self.create_view.as_view(), name="voucher-create"),
            path("update/<int:pk>/", self.update_view.as_view(), name="voucher-update"),
            path("delete/<int:pk>/", self.delete_view.as_view(), name="voucher-delete"),
            path("stats/<int:pk>/", self.stats_view.as_view(), name="voucher-stats"),
            path("sets/", self.set_list_view.as_view(), name="voucher-set-list"),
            path(
                "sets/create/",
                self.set_create_view.as_view(),
                name="voucher-set-create",
            ),
            path(
                "sets/update/<int:pk>/",
                self.set_update_view.as_view(),
                name="voucher-set-update",
            ),
            path(
                "sets/detail/<int:pk>/",
                self.set_detail_view.as_view(),
                name="voucher-set-detail",
            ),
            path(
                "sets/download/<int:pk>/",
                self.set_download_view.as_view(),
                name="voucher-set-download",
            ),
            path(
                "sets/delete/<int:pk>/",
                self.set_delete_view.as_view(),
                name="voucher-set-delete",
            ),
        ]
        return self.post_process_urls(urls)



from decimal import Decimal

from django.db import models
from django.utils.translation import gettext_lazy as _

from oscar.core.compat import AUTH_USER_MODEL


class AbstractProductRecord(models.Model):
    """
    A record of a how popular a product is.

    This used be auto-merchandising to display the most popular
    products.
    """

    product = models.OneToOneField(
        "catalogue.Product",
        verbose_name=_("Product"),
        related_name="stats",
        on_delete=models.CASCADE,
    )

    # Data used for generating a score
    num_views = models.PositiveIntegerField(_("Views"), default=0)
    num_basket_additions = models.PositiveIntegerField(_("Basket Additions"), default=0)
    num_purchases = models.PositiveIntegerField(
        _("Purchases"), default=0, db_index=True
    )

    # Product score - used within search
    score = models.FloatField(_("Score"), default=0.00)

    class Meta:
        abstract = True
        app_label = "analytics"
        ordering = ["-num_purchases"]
        verbose_name = _("Product record")
        verbose_name_plural = _("Product records")

    def __str__(self):
        return _("Record for '%s'") % self.product


class AbstractUserRecord(models.Model):
    """
    A record of a user's activity.
    """

    user = models.OneToOneField(
        AUTH_USER_MODEL, verbose_name=_("User"), on_delete=models.CASCADE
    )

    # Browsing stats
    num_product_views = models.PositiveIntegerField(_("Product Views"), default=0)
    num_basket_additions = models.PositiveIntegerField(_("Basket Additions"), default=0)

    # Order stats
    num_orders = models.PositiveIntegerField(_("Orders"), default=0, db_index=True)
    num_order_lines = models.PositiveIntegerField(
        _("Order Lines"), default=0, db_index=True
    )
    num_order_items = models.PositiveIntegerField(
        _("Order Items"), default=0, db_index=True
    )
    total_spent = models.DecimalField(
        _("Total Spent"), decimal_places=2, max_digits=12, default=Decimal("0.00")
    )
    date_last_order = models.DateTimeField(_("Last Order Date"), blank=True, null=True)

    class Meta:
        abstract = True
        app_label = "analytics"
        verbose_name = _("User record")
        verbose_name_plural = _("User records")


class AbstractUserProductView(models.Model):
    user = models.ForeignKey(
        AUTH_USER_MODEL, verbose_name=_("User"), on_delete=models.CASCADE
    )
    product = models.ForeignKey(
        "catalogue.Product", on_delete=models.CASCADE, verbose_name=_("Product")
    )
    date_created = models.DateTimeField(_("Date Created"), auto_now_add=True)

    class Meta:
        abstract = True
        app_label = "analytics"
        ordering = ["-pk"]
        verbose_name = _("User product view")
        verbose_name_plural = _("User product views")

    def __str__(self):
        return _("%(user)s viewed '%(product)s'") % {
            "user": self.user,
            "product": self.product,
        }


class AbstractUserSearch(models.Model):
    user = models.ForeignKey(
        AUTH_USER_MODEL, on_delete=models.CASCADE, verbose_name=_("User")
    )
    query = models.CharField(_("Search term"), max_length=255, db_index=True)
    date_created = models.DateTimeField(_("Date Created"), auto_now_add=True)

    class Meta:
        abstract = True
        app_label = "analytics"
        ordering = ["-pk"]
        verbose_name = _("User search query")
        verbose_name_plural = _("User search queries")

    def __str__(self):
        return _("%(user)s searched for '%(query)s'") % {
            "user": self.user,
            "query": self.query,
        }



# Generated by Django 2.1.5 on 2019-02-11 11:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0014_auto_20181115_1953'),
    ]

    operations = [
        migrations.AddField(
            model_name='product',
            name='is_public',
            field=models.BooleanField(default=True, help_text='Show this product in search results and catalogue listings.', verbose_name='Is public'),
        ),
    ]



from django import template

register = template.Library()


@register.simple_tag
def shipping_charge(method, basket):
    """
    Template tag for calculating the shipping charge for a given shipping
    method and basket, and injecting it into the template context.
    """
    return method.calculate(basket)


@register.simple_tag
def shipping_charge_discount(method, basket):
    """
    Template tag for calculating the shipping discount for a given shipping
    method and basket, and injecting it into the template context.
    """
    return method.discount(basket)


@register.simple_tag
def shipping_charge_excl_discount(method, basket):
    """
    Template tag for calculating the shipping charge (excluding discounts) for
    a given shipping method and basket, and injecting it into the template
    context.
    """
    return method.calculate_excl_discount(basket)



from oscar.core import prices
from django.conf import settings


class Base(object):
    """
    The interface that any pricing policy must support
    """

    #: Whether any prices exist
    exists = False

    #: Whether tax is known
    is_tax_known = False

    #: Price excluding tax
    excl_tax = None

    #: Price including tax
    incl_tax = None

    #: Price to use for offer calculations
    @property
    def effective_price(self):
        if settings.OSCAR_OFFERS_INCL_TAX:
            return self.incl_tax
        # Default to using the price excluding tax for calculations
        return self.excl_tax

    #: Price tax
    tax = None

    # Code used to store the vat rate reference
    tax_code = None

    #: Retail price
    retail = None

    #: Price currency (3 char code)
    currency = None

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.__dict__)


class Unavailable(Base):
    """
    This should be used as a pricing policy when a product is unavailable and
    no prices are known.
    """


class FixedPrice(Base):
    """
    This should be used for when the price of a product is known in advance.

    It can work for when tax isn't known (like in the US).

    Note that this price class uses the tax-exclusive price for offers, even if
    the tax is known.  This may not be what you want.  Use the
    TaxInclusiveFixedPrice class if you want offers to use tax-inclusive
    prices.
    """

    exists = True

    def __init__(self, currency, excl_tax, tax=None, tax_code=None):
        super().__init__()
        self.currency = currency
        self.excl_tax = excl_tax
        self.tax = tax
        self.tax_code = tax_code

    @property
    def incl_tax(self):
        if self.is_tax_known:
            return self.excl_tax + self.tax
        raise prices.TaxNotKnown("Can't calculate price.incl_tax as tax isn't known")

    @property
    def is_tax_known(self) -> bool:
        """
        Test whether the tax is known or not
        """
        return self.tax is not None


class TaxInclusiveFixedPrice(FixedPrice):
    """
    Specialised version of FixedPrice that must have tax passed.  It also
    specifies that offers should use the tax-inclusive price (which is the norm
    in the UK).
    """

    exists = is_tax_known = True

    @property
    def incl_tax(self):
        return self.excl_tax + self.tax

    @property
    def effective_price(self):
        return self.incl_tax



# pylint: disable=attribute-defined-outside-init
import datetime
from decimal import Decimal as D
from decimal import InvalidOperation

from django.conf import settings
from django.contrib import messages
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Count, Q, Sum, fields
from django.http import Http404, HttpResponse, HttpResponseRedirect
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import DetailView, FormView, ListView, UpdateView

from oscar.apps.order import exceptions as order_exceptions
from oscar.apps.payment.exceptions import PaymentError
from oscar.core.compat import UnicodeCSVWriter
from oscar.core.loading import get_class, get_model
from oscar.core.utils import datetime_combine, format_datetime
from oscar.views import sort_queryset
from oscar.views.generic import BulkEditMixin

Partner = get_model("partner", "Partner")
Transaction = get_model("payment", "Transaction")
SourceType = get_model("payment", "SourceType")
Order = get_model("order", "Order")
OrderNote = get_model("order", "OrderNote")
ShippingAddress = get_model("order", "ShippingAddress")
Line = get_model("order", "Line")
ShippingEventType = get_model("order", "ShippingEventType")
PaymentEventType = get_model("order", "PaymentEventType")
EventHandlerMixin = get_class("order.mixins", "EventHandlerMixin")
OrderStatsForm = get_class("dashboard.orders.forms", "OrderStatsForm")
OrderSearchForm = get_class("dashboard.orders.forms", "OrderSearchForm")
OrderNoteForm = get_class("dashboard.orders.forms", "OrderNoteForm")
ShippingAddressForm = get_class("dashboard.orders.forms", "ShippingAddressForm")
OrderStatusForm = get_class("dashboard.orders.forms", "OrderStatusForm")


def queryset_orders_for_user(user):
    """
    Returns a queryset of all orders that a user is allowed to access.
    A staff user may access all orders.
    To allow access to an order for a non-staff user, at least one line's
    partner has to have the user in the partner's list.
    """
    queryset = Order._default_manager.select_related(
        "billing_address",
        "billing_address__country",
        "shipping_address",
        "shipping_address__country",
        "user",
    ).prefetch_related("lines", "status_changes")
    if user.is_staff:
        return queryset
    else:
        partners = Partner._default_manager.filter(users=user)
        return queryset.filter(lines__partner__in=partners).distinct()


def get_order_for_user_or_404(user, number):
    try:
        return queryset_orders_for_user(user).get(number=number)
    except ObjectDoesNotExist:
        raise Http404()


class OrderStatsView(FormView):
    """
    Dashboard view for order statistics.
    Supports the permission-based dashboard.
    """

    template_name = "oscar/dashboard/orders/statistics.html"
    form_class = OrderStatsForm

    def get(self, request, *args, **kwargs):
        return self.post(request, *args, **kwargs)

    def form_valid(self, form):
        ctx = self.get_context_data(form=form, filters=form.get_filters())
        return self.render_to_response(ctx)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["data"] = self.request.GET
        return kwargs

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        filters = kwargs.get("filters", {})
        ctx.update(self.get_stats(filters))
        ctx["title"] = kwargs["form"].get_filter_description()
        return ctx

    def get_stats(self, filters):
        orders = queryset_orders_for_user(self.request.user).filter(**filters)
        stats = {
            "total_orders": orders.count(),
            "total_lines": Line.objects.filter(order__in=orders).count(),
            "total_revenue": orders.aggregate(Sum("total_incl_tax"))[
                "total_incl_tax__sum"
            ]
            or D("0.00"),
            "order_status_breakdown": orders.order_by("status")
            .values("status")
            .annotate(freq=Count("id")),
        }
        return stats


class OrderListView(EventHandlerMixin, BulkEditMixin, ListView):
    """
    Dashboard view for a list of orders.
    Supports the permission-based dashboard.
    """

    model = Order
    context_object_name = "orders"
    template_name = "oscar/dashboard/orders/order_list.html"
    form_class = OrderSearchForm
    paginate_by = settings.OSCAR_DASHBOARD_ITEMS_PER_PAGE
    actions = ("download_selected_orders", "change_order_statuses")
    CSV_COLUMNS = {
        "number": _("Order number"),
        "value": _("Order value"),
        "date": _("Date of purchase"),
        "num_items": _("Number of items"),
        "status": _("Order status"),
        "customer": _("Customer email address"),
        "shipping_address_name": _("Deliver to name"),
        "billing_address_name": _("Bill to name"),
    }

    def dispatch(self, request, *args, **kwargs):
        # base_queryset is equal to all orders the user is allowed to access
        self.base_queryset = queryset_orders_for_user(request.user).order_by(
            "-date_placed"
        )
        return super().dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        if (
            "order_number" in request.GET
            and request.GET.get("response_format", "html") == "html"
        ):
            # Redirect to Order detail page if valid order number is given
            try:
                order = self.base_queryset.get(number=request.GET["order_number"])
            except Order.DoesNotExist:
                pass
            else:
                return redirect("dashboard:order-detail", number=order.number)
        return super().get(request, *args, **kwargs)

    def get_queryset(self):
        """
        Build the queryset for this list.
        """
        queryset = sort_queryset(
            self.base_queryset, self.request, ["number", "total_incl_tax"]
        )

        self.form = self.form_class(self.request.GET)
        if not self.form.is_valid():
            return queryset

        data = self.form.cleaned_data

        if data["order_number"]:
            queryset = self.base_queryset.filter(
                number__istartswith=data["order_number"]
            )

        if data["name"]:
            # If the value is two words, then assume they are first name and
            # last name
            parts = data["name"].split()
            allow_anon = getattr(settings, "OSCAR_ALLOW_ANON_CHECKOUT", False)

            if len(parts) == 1:
                parts = [data["name"], data["name"]]
            else:
                parts = [parts[0], parts[1:]]

            query = Q(user__first_name__istartswith=parts[0])
            query |= Q(user__last_name__istartswith=parts[1])
            if allow_anon:
                query |= Q(billing_address__first_name__istartswith=parts[0])
                query |= Q(shipping_address__first_name__istartswith=parts[0])
                query |= Q(billing_address__last_name__istartswith=parts[1])
                query |= Q(shipping_address__last_name__istartswith=parts[1])

            queryset = queryset.filter(query).distinct()

        if data["product_title"]:
            queryset = queryset.filter(
                lines__title__istartswith=data["product_title"]
            ).distinct()

        if data["upc"]:
            queryset = queryset.filter(lines__upc=data["upc"])

        if data["partner_sku"]:
            queryset = queryset.filter(lines__partner_sku=data["partner_sku"])

        if data["date_from"] and data["date_to"]:
            date_to = datetime_combine(data["date_to"], datetime.time.max)
            date_from = datetime_combine(data["date_from"], datetime.time.min)
            queryset = queryset.filter(
                date_placed__gte=date_from, date_placed__lt=date_to
            )
        elif data["date_from"]:
            date_from = datetime_combine(data["date_from"], datetime.time.min)
            queryset = queryset.filter(date_placed__gte=date_from)
        elif data["date_to"]:
            date_to = datetime_combine(data["date_to"], datetime.time.max)
            queryset = queryset.filter(date_placed__lt=date_to)

        if data["voucher"]:
            queryset = queryset.filter(
                discounts__voucher_code=data["voucher"]
            ).distinct()

        if data["payment_method"]:
            queryset = queryset.filter(
                sources__source_type__code=data["payment_method"]
            ).distinct()

        if data["status"]:
            queryset = queryset.filter(status=data["status"])

        return queryset

    def get_search_filter_descriptions(self):
        """Describe the filters used in the search.

        These are user-facing messages describing what filters
        were used to filter orders in the search query.

        Returns:
            list of unicode messages

        """
        descriptions = []

        # Attempt to retrieve data from the submitted form
        # If the form hasn't been submitted, then `cleaned_data`
        # won't be set, so default to None.
        data = getattr(self.form, "cleaned_data", None)

        if data is None:
            return descriptions

        if data.get("order_number"):
            descriptions.append(
                _('Order number starts with "{order_number}"').format(
                    order_number=data["order_number"]
                )
            )

        if data.get("name"):
            descriptions.append(
                _('Customer name matches "{customer_name}"').format(
                    customer_name=data["name"]
                )
            )

        if data.get("product_title"):
            descriptions.append(
                _('Product name matches "{product_name}"').format(
                    product_name=data["product_title"]
                )
            )

        if data.get("upc"):
            descriptions.append(
                # Translators: "UPC" means "universal product code" and it is
                # used to uniquely identify a product in an online store.
                # "Item" in this context means an item in an order placed
                # in an online store.
                _('Includes an item with UPC "{upc}"').format(upc=data["upc"])
            )

        if data.get("partner_sku"):
            descriptions.append(
                # Translators: "SKU" means "stock keeping unit" and is used to
                # identify products that can be shipped from an online store.
                # A "partner" is a company that ships items to users who
                # buy things in an online store.
                _('Includes an item with partner SKU "{partner_sku}"').format(
                    partner_sku=data["partner_sku"]
                )
            )

        if data.get("date_from") and data.get("date_to"):
            descriptions.append(
                # Translators: This string refers to orders in an online
                # store that were made within a particular date range.
                _("Placed between {start_date} and {end_date}").format(
                    start_date=data["date_from"], end_date=data["date_to"]
                )
            )

        elif data.get("date_from"):
            descriptions.append(
                # Translators: This string refers to orders in an online store
                # that were made after a particular date.
                _("Placed after {start_date}").format(start_date=data["date_from"])
            )

        elif data.get("date_to"):
            end_date = data["date_to"] + datetime.timedelta(days=1)
            descriptions.append(
                # Translators: This string refers to orders in an online store
                # that were made before a particular date.
                _("Placed before {end_date}").format(end_date=end_date)
            )

        if data.get("voucher"):
            descriptions.append(
                # Translators: A "voucher" is a coupon that can be applied to
                # an order in an online store in order to receive a discount.
                # The voucher "code" is a string that users can enter to
                # receive the discount.
                _('Used voucher code "{voucher_code}"').format(
                    voucher_code=data["voucher"]
                )
            )

        if data.get("payment_method"):
            payment_type = SourceType.objects.get(code=data["payment_method"])
            descriptions.append(
                # Translators: A payment method is a way of paying for an
                # item in an online store.  For example, a user can pay
                # with a credit card or PayPal.
                _("Paid using {payment_method}").format(
                    payment_method=payment_type.name
                )
            )

        if data.get("status"):
            descriptions.append(
                # Translators: This string refers to an order in an
                # online store.  Some examples of order status are
                # "purchased", "cancelled", or "refunded".
                _("Order status is {order_status}").format(order_status=data["status"])
            )

        return descriptions

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["form"] = self.form
        ctx["order_statuses"] = Order.all_statuses()
        ctx["search_filters"] = self.get_search_filter_descriptions()
        return ctx

    def is_csv_download(self):
        return self.request.GET.get("response_format", None) == "csv"

    def get_paginate_by(self, queryset):
        return None if self.is_csv_download() else self.paginate_by

    def render_to_response(self, context, **response_kwargs):
        if self.is_csv_download():
            return self.download_selected_orders(self.request, context["object_list"])
        return super().render_to_response(context, **response_kwargs)

    def get_download_filename(self, request):
        return "orders.csv"

    def get_row_values(self, order):
        row = {
            "number": order.number,
            "customer": order.email,
            "num_items": order.num_items,
            "date": format_datetime(order.date_placed, "DATETIME_FORMAT"),
            "value": order.total_incl_tax,
            "status": order.status,
        }
        if order.shipping_address:
            row["shipping_address_name"] = order.shipping_address.name
        if order.billing_address:
            row["billing_address_name"] = order.billing_address.name
        return row

    def download_selected_orders(self, request, orders):
        response = HttpResponse(content_type="text/csv")
        response["Content-Disposition"] = (
            "attachment; filename=%s" % self.get_download_filename(request)
        )
        writer = UnicodeCSVWriter(open_file=response)

        writer.writerow(self.CSV_COLUMNS.values())
        for order in orders:
            row_values = self.get_row_values(order)
            writer.writerow([row_values.get(column, "") for column in self.CSV_COLUMNS])
        return response

    def change_order_statuses(self, request, orders):
        for order in orders:
            self.change_order_status(request, order)
        return redirect("dashboard:order-list")

    def change_order_status(self, request, order):
        # This method is pretty similar to what
        # OrderDetailView.change_order_status does. Ripe for refactoring.
        new_status = request.POST["new_status"].strip()
        if not new_status:
            messages.error(request, _("The new status '%s' is not valid") % new_status)
        elif new_status not in order.available_statuses():
            messages.error(
                request,
                _("The new status '%s' is not valid for this order") % new_status,
            )
        else:
            handler = self.get_handler(user=request.user)
            old_status = order.status
            try:
                handler.handle_order_status_change(order, new_status)
            except PaymentError as e:
                messages.error(
                    request,
                    _("Unable to change order status due to payment error: %s") % e,
                )
            else:
                msg = _(
                    "Order status changed from '%(old_status)s' to '%(new_status)s'"
                ) % {"old_status": old_status, "new_status": new_status}
                messages.info(request, msg)
                order.notes.create(
                    user=request.user, message=msg, note_type=OrderNote.SYSTEM
                )


class OrderDetailView(EventHandlerMixin, DetailView):
    """
    Dashboard view to display a single order.

    Supports the permission-based dashboard.
    """

    model = Order
    context_object_name = "order"
    template_name = "oscar/dashboard/orders/order_detail.html"

    # These strings are method names that are allowed to be called from a
    # submitted form.
    order_actions = (
        "save_note",
        "delete_note",
        "change_order_status",
        "create_order_payment_event",
    )
    line_actions = (
        "change_line_statuses",
        "create_shipping_event",
        "create_payment_event",
    )

    def get_object(self, queryset=None):
        return get_order_for_user_or_404(self.request.user, self.kwargs["number"])

    def get_order_lines(self):
        return self.object.lines.all()

    def post(self, request, *args, **kwargs):
        # For POST requests, we use a dynamic dispatch technique where a
        # parameter specifies what we're trying to do with the form submission.
        # We distinguish between order-level actions and line-level actions.

        order = self.object = self.get_object()

        # Look for order-level action first
        if "order_action" in request.POST:
            return self.handle_order_action(
                request, order, request.POST["order_action"]
            )

        # Look for line-level action
        if "line_action" in request.POST:
            return self.handle_line_action(request, order, request.POST["line_action"])

        return self.reload_page(error=_("No valid action submitted"))

    def handle_order_action(self, request, order, action):
        if action not in self.order_actions:
            return self.reload_page(error=_("Invalid action"))
        return getattr(self, action)(request, order)

    def handle_line_action(self, request, order, action):
        if action not in self.line_actions:
            return self.reload_page(error=_("Invalid action"))

        # Load requested lines
        line_ids = request.POST.getlist("selected_line")
        if len(line_ids) == 0:
            return self.reload_page(error=_("You must select some lines to act on"))

        lines = self.get_order_lines()
        lines = lines.filter(id__in=line_ids)
        if len(line_ids) != len(lines):
            return self.reload_page(error=_("Invalid lines requested"))

        # Build list of line quantities
        line_quantities = []
        for line in lines:
            qty = request.POST.get("selected_line_qty_%s" % line.id)
            try:
                qty = int(qty)
            except ValueError:
                qty = None
            if qty is None or qty <= 0:
                error_msg = _("The entered quantity for line #%s is not valid")
                return self.reload_page(error=error_msg % line.id)
            elif qty > line.quantity:
                error_msg = _(
                    "The entered quantity for line #%(line_id)s "
                    "should not be higher than %(quantity)s"
                )
                kwargs = {"line_id": line.id, "quantity": line.quantity}
                return self.reload_page(error=error_msg % kwargs)

            line_quantities.append(qty)

        return getattr(self, action)(request, order, lines, line_quantities)

    def reload_page(self, fragment=None, error=None):
        url = reverse("dashboard:order-detail", kwargs={"number": self.object.number})
        if fragment:
            url += "#" + fragment
        if error:
            messages.error(self.request, error)
        return HttpResponseRedirect(url)

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["active_tab"] = kwargs.get("active_tab", "lines")

        # Forms
        ctx["note_form"] = self.get_order_note_form()
        ctx["order_status_form"] = self.get_order_status_form()

        ctx["lines"] = self.get_order_lines()
        ctx["line_statuses"] = Line.all_statuses()
        ctx["shipping_event_types"] = ShippingEventType.objects.all()
        ctx["payment_event_types"] = PaymentEventType.objects.all()

        ctx["payment_transactions"] = self.get_payment_transactions()

        return ctx

    # Data fetching methods for template context

    def get_payment_transactions(self):
        return Transaction.objects.filter(source__order=self.object)

    def get_order_note_form(self):
        kwargs = {"order": self.object, "user": self.request.user, "data": None}
        if self.request.method == "POST":
            kwargs["data"] = self.request.POST
        note_id = self.kwargs.get("note_id", None)
        if note_id:
            note = get_object_or_404(OrderNote, order=self.object, id=note_id)
            if note.is_editable():
                kwargs["instance"] = note
        return OrderNoteForm(**kwargs)

    def get_order_status_form(self):
        data = None
        if self.request.method == "POST":
            data = self.request.POST
        return OrderStatusForm(order=self.object, data=data)

    # Order-level actions

    # pylint: disable=unused-argument
    def save_note(self, request, order):
        form = self.get_order_note_form()
        if form.is_valid():
            form.save()
            messages.success(self.request, _("Note saved"))
            return self.reload_page(fragment="notes")

        ctx = self.get_context_data(note_form=form, active_tab="notes")
        return self.render_to_response(ctx)

    def delete_note(self, request, order):
        try:
            note = order.notes.get(id=request.POST.get("note_id", None))
        except ObjectDoesNotExist:
            messages.error(request, _("Note cannot be deleted"))
        else:
            messages.info(request, _("Note deleted"))
            note.delete()
        return self.reload_page()

    def change_order_status(self, request, order):
        form = self.get_order_status_form()
        if not form.is_valid():
            return self.reload_page(error=_("Invalid form submission"))

        old_status, new_status = order.status, form.cleaned_data["new_status"]
        handler = self.get_handler(user=request.user)

        success_msg = _(
            "Order status changed from '%(old_status)s' to '%(new_status)s'"
        ) % {"old_status": old_status, "new_status": new_status}
        try:
            handler.handle_order_status_change(order, new_status, note_msg=success_msg)
        except PaymentError as e:
            messages.error(
                request,
                _("Unable to change order status due to payment error: %s") % e,
            )
        except order_exceptions.InvalidOrderStatus:
            # The form should validate against this, so we should only end up
            # here during race conditions.
            messages.error(
                request,
                _(
                    "Unable to change order status as the requested "
                    "new status is not valid"
                ),
            )
        else:
            messages.info(request, success_msg)
        return self.reload_page()

    def create_order_payment_event(self, request, order):
        """
        Create a payment event for the whole order
        """
        amount_str = request.POST.get("amount", None)
        try:
            amount = D(amount_str)
        except InvalidOperation:
            messages.error(request, _("Please choose a valid amount"))
            return self.reload_page()
        return self._create_payment_event(request, order, amount)

    # Line-level actions

    # pylint: disable=unused-argument
    def change_line_statuses(self, request, order, lines, quantities):
        new_status = request.POST["new_status"].strip()
        if not new_status:
            messages.error(request, _("The new status '%s' is not valid") % new_status)
            return self.reload_page()
        errors = []
        for line in lines:
            if new_status not in line.available_statuses():
                errors.append(
                    _("'%(status)s' is not a valid new status for line %(line_id)d")
                    % {"status": new_status, "line_id": line.id}
                )
        if errors:
            messages.error(request, "\n".join(errors))
            return self.reload_page()

        msgs = []
        for line in lines:
            msg = _(
                "Status of line #%(line_id)d changed from '%(old_status)s'"
                " to '%(new_status)s'"
            ) % {
                "line_id": line.id,
                "old_status": line.status,
                "new_status": new_status,
            }
            msgs.append(msg)
            line.set_status(new_status)
        message = "\n".join(msgs)
        messages.info(request, message)
        order.notes.create(
            user=request.user, message=message, note_type=OrderNote.SYSTEM
        )
        return self.reload_page()

    def create_shipping_event(self, request, order, lines, quantities):
        code = request.POST["shipping_event_type"]
        try:
            event_type = ShippingEventType._default_manager.get(code=code)
        except ShippingEventType.DoesNotExist:
            messages.error(request, _("The event type '%s' is not valid") % code)
            return self.reload_page()

        reference = request.POST.get("reference", None)
        try:
            self.get_handler().handle_shipping_event(
                order, event_type, lines, quantities, reference=reference
            )
        except order_exceptions.InvalidShippingEvent as e:
            messages.error(request, _("Unable to create shipping event: %s") % e)
        except order_exceptions.InvalidStatus as e:
            messages.error(request, _("Unable to create shipping event: %s") % e)
        except PaymentError as e:
            messages.error(
                request,
                _("Unable to create shipping event due to payment error: %s") % e,
            )
        else:
            messages.success(request, _("Shipping event created"))
        return self.reload_page()

    def create_payment_event(self, request, order, lines, quantities):
        """
        Create a payment event for a subset of order lines
        """
        amount_str = request.POST.get("amount", None)

        # If no amount passed, then we add up the total of the selected lines
        if not amount_str:
            amount = sum([line.line_price_incl_tax for line in lines])
        else:
            try:
                amount = D(amount_str)
            except InvalidOperation:
                messages.error(request, _("Please choose a valid amount"))
                return self.reload_page()

        return self._create_payment_event(request, order, amount, lines, quantities)

    def _create_payment_event(
        self, request, order, amount, lines=None, quantities=None
    ):
        code = request.POST.get("payment_event_type")
        try:
            event_type = PaymentEventType._default_manager.get(code=code)
        except PaymentEventType.DoesNotExist:
            messages.error(request, _("The event type '%s' is not valid") % code)
            return self.reload_page()
        try:
            self.get_handler().handle_payment_event(
                order, event_type, amount, lines, quantities
            )
        except PaymentError as e:
            messages.error(
                request,
                _("Unable to create payment event due to payment error: %s") % e,
            )
        except order_exceptions.InvalidPaymentEvent as e:
            messages.error(request, _("Unable to create payment event: %s") % e)
        else:
            messages.info(request, _("Payment event created"))
        return self.reload_page()


class LineDetailView(DetailView):
    """
    Dashboard view to show a single line of an order.
    Supports the permission-based dashboard.
    """

    model = Line
    context_object_name = "line"
    template_name = "oscar/dashboard/orders/line_detail.html"

    def get_object(self, queryset=None):
        order = get_order_for_user_or_404(self.request.user, self.kwargs["number"])
        try:
            return order.lines.get(pk=self.kwargs["line_id"])
        except self.model.DoesNotExist:
            raise Http404()

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["order"] = self.object.order
        return ctx


def get_changes_between_models(model1, model2, excludes=None):
    """
    Return a dict of differences between two model instances
    """
    if excludes is None:
        excludes = []
    changes = {}
    for field in model1._meta.fields:
        if (
            isinstance(field, (fields.AutoField, fields.related.RelatedField))
            or field.name in excludes
        ):
            continue

        if field.value_from_object(model1) != field.value_from_object(model2):
            changes[field.verbose_name] = (
                field.value_from_object(model1),
                field.value_from_object(model2),
            )
    return changes


def get_change_summary(model1, model2):
    """
    Generate a summary of the changes between two address models
    """
    changes = get_changes_between_models(model1, model2, ["search_text"])
    change_descriptions = []
    for field, delta in changes.items():
        change_descriptions.append(
            _("%(field)s changed from '%(old_value)s' to '%(new_value)s'")
            % {"field": field, "old_value": delta[0], "new_value": delta[1]}
        )
    return "\n".join(change_descriptions)


class ShippingAddressUpdateView(UpdateView):
    """
    Dashboard view to update an order's shipping address.
    Supports the permission-based dashboard.
    """

    model = ShippingAddress
    context_object_name = "address"
    template_name = "oscar/dashboard/orders/shippingaddress_form.html"
    form_class = ShippingAddressForm

    def get_object(self, queryset=None):
        order = get_order_for_user_or_404(self.request.user, self.kwargs["number"])
        return get_object_or_404(self.model, order=order)

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["order"] = self.object.order
        return ctx

    def form_valid(self, form):
        old_address = ShippingAddress.objects.get(id=self.object.id)
        response = super().form_valid(form)
        changes = get_change_summary(old_address, self.object)
        if changes:
            msg = _("Delivery address updated:\n%s") % changes
            self.object.order.notes.create(
                user=self.request.user, message=msg, note_type=OrderNote.SYSTEM
            )
        return response

    def get_success_url(self):
        messages.info(self.request, _("Delivery address updated"))
        return reverse(
            "dashboard:order-detail",
            kwargs={
                "number": self.object.order.number,
            },
        )



# -*- coding: utf-8 -*-
from django import forms

from oscar.core.loading import get_model

WishList = get_model("wishlists", "WishList")
Line = get_model("wishlists", "Line")
WishListSharedEmail = get_model("wishlists", "WishListSharedEmail")


class WishListForm(forms.ModelForm):
    def __init__(self, user, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instance.owner = user

    class Meta:
        model = WishList
        fields = ("name", "visibility")


class WishListLineForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["quantity"].widget.attrs["class"] = "input-mini"


class WishListSharedEmailForm(forms.ModelForm):
    class Meta:
        model = WishListSharedEmail
        fields = ("email",)



import factory

from oscar.core.loading import get_model

__all__ = ["WishListFactory"]


class WishListFactory(factory.django.DjangoModelFactory):
    name = factory.Sequence(lambda n: "Wishlist %d" % n)

    class Meta:
        model = get_model("wishlists", "WishList")



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-04-23 11:17
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('offer', '0004_auto_20170415_1518'),
    ]

    operations = [
        migrations.AlterField(
            model_name='conditionaloffer',
            name='benefit',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='offers', to='offer.Benefit', verbose_name='Benefit'),
        ),
        migrations.AlterField(
            model_name='conditionaloffer',
            name='condition',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='offers', to='offer.Condition', verbose_name='Condition'),
        ),
    ]



# Generated by Django 2.2.10 on 2020-07-24 08:09

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('order', '0009_surcharge'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='line',
            name='est_dispatch_date',
        ),
        migrations.RemoveField(
            model_name='line',
            name='unit_cost_price',
        ),
        migrations.RemoveField(
            model_name='line',
            name='unit_retail_price',
        ),
    ]



import os
import queue
import shutil
import threading
from datetime import date

from django.conf import settings
from django.contrib.auth.models import AnonymousUser
from django.contrib.messages.storage.fallback import FallbackStorage
from django.contrib.sessions.backends.db import SessionStore
from django.core import mail
from django.core.signing import Signer
from django.db import connection
from django.test import RequestFactory as BaseRequestFactory
from sorl.thumbnail.conf import settings as sorl_settings

from oscar.core.loading import get_class, get_model
from oscar.core.thumbnails import get_thumbnailer
from oscar.test.factories import ProductImageFactory, UserFactory

OSCAR_IMAGE_FOLDER_FORMATTED = "images/products/{0}/{1:02d}/".format(
    date.today().year, date.today().month
)
FULL_PATH_TO_IMAGES_FOLDER = os.path.join(
    settings.MEDIA_ROOT, OSCAR_IMAGE_FOLDER_FORMATTED
)
FULL_PATH_TO_SORL_THUMBNAILS_FOLDER = os.path.join(
    settings.MEDIA_ROOT, sorl_settings.THUMBNAIL_PREFIX
)
EASY_THUMBNAIL_BASEDIR = "thumbnails"
FULL_PATH_TO_EASY_THUMBNAILS_FOLDER = os.path.join(
    settings.MEDIA_ROOT, EASY_THUMBNAIL_BASEDIR, OSCAR_IMAGE_FOLDER_FORMATTED
)


def remove_image_folders():
    if os.path.exists(FULL_PATH_TO_IMAGES_FOLDER):
        shutil.rmtree(FULL_PATH_TO_IMAGES_FOLDER)
    if os.path.exists(FULL_PATH_TO_SORL_THUMBNAILS_FOLDER):
        shutil.rmtree(FULL_PATH_TO_SORL_THUMBNAILS_FOLDER)
    if os.path.exists(FULL_PATH_TO_EASY_THUMBNAILS_FOLDER):
        shutil.rmtree(FULL_PATH_TO_EASY_THUMBNAILS_FOLDER)


def get_thumbnail_full_path(thumbnail_url):
    # Thumbnail URL looks like "/media/...........jpg". To be able to
    # create full path we need to remove the first "/" from it.
    return os.path.join(settings.PUBLIC_ROOT, thumbnail_url[1:])


class ThumbnailMixin:
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        # Remove images folders - they could be created in the previous tests.
        remove_image_folders()

    def setUp(self):
        super().setUp()
        self.thumbnail_options = {
            "size": "x50",
            "upscale": False,
        }

    def tearDown(self):
        super().tearDown()
        # Remove created images folders after each test.
        remove_image_folders()

    def create_product_images(self, qty=2, product=None):
        self._test_images_folder_is_empty()

        kwargs = {}
        if product is not None:
            kwargs["product"] = product
        self.images = ProductImageFactory.create_batch(qty, **kwargs)

        file_names = os.listdir(FULL_PATH_TO_IMAGES_FOLDER)
        assert len(file_names) == qty  # New images added.

    def create_thumbnails(self):
        thumbnailer = get_thumbnailer()
        thumbnails_full_paths = []

        for image in self.images:
            thumbnail = thumbnailer.generate_thumbnail(
                image.original, **self.thumbnail_options
            )
            thumbnails_full_paths.append(get_thumbnail_full_path(thumbnail.url))

        # Check thumbnails exist in the file system.
        for path in thumbnails_full_paths:
            assert os.path.isfile(
                path
            )  # `isfile` returns `True` if path is an existing regular file.

        return thumbnails_full_paths

    def _test_images_folder_is_empty(self):
        if not os.path.exists(FULL_PATH_TO_IMAGES_FOLDER):
            os.makedirs(FULL_PATH_TO_IMAGES_FOLDER)

        file_names = os.listdir(FULL_PATH_TO_IMAGES_FOLDER)
        assert len(file_names) == 0

    def _test_thumbnails_not_exist(self, thumbnails_full_paths):
        for path in thumbnails_full_paths:
            assert not os.path.isfile(path)


class EmailsMixin:
    DJANGO_IMPROPERLY_CONFIGURED_MSG = (
        'You\'re using the Django "sites framework" '
        "without having set the SITE_ID setting. Create a site in your database and set "
        "the SITE_ID setting or pass a request to Site.objects.get_current() to fix this error."
    )

    def setUp(self):
        super().setUp()
        self.user = UserFactory()
        factory = RequestFactory(SERVER_NAME="example.com")
        self.request = factory.get("/")

    def _test_send_plain_text_and_html(self, outboxed_email):
        email = outboxed_email

        assert "</p>" not in email.body  # Plain text body (because w/o </p> tags)

        html_content = email.alternatives[0][0]
        assert "</p>" in html_content

        mimetype = email.alternatives[0][1]
        assert mimetype == "text/html"

    def _test_common_part(self):
        assert len(mail.outbox) == 1
        assert mail.outbox[0].to == [self.user.email]
        self._test_send_plain_text_and_html(mail.outbox[0])


class RequestFactory(BaseRequestFactory):
    Basket = get_model("basket", "basket")
    selector = get_class("partner.strategy", "Selector")()

    def request(self, user=None, basket=None, **request):
        request = super().request(**request)
        request.user = user or AnonymousUser()
        request.session = SessionStore()
        request._messages = FallbackStorage(request)

        # Mimic basket middleware
        # pylint: disable=E1101
        request.strategy = self.selector.strategy(request=request, user=request.user)
        request.basket = basket or self.Basket()
        request.basket.strategy = request.strategy
        request.basket_hash = Signer().sign(basket.pk) if basket else None
        request.cookies_to_delete = []

        return request


def run_concurrently(fn, kwargs=None, num_threads=5):
    exceptions = queue.Queue()

    def worker(**kwargs):
        try:
            fn(**kwargs)
        except Exception as exc:
            exceptions.put(exc)
        else:
            exceptions.put(None)
        finally:
            connection.close()

    kwargs = kwargs if kwargs is not None else {}

    # Run them
    threads = [
        threading.Thread(target=worker, name="thread-%d" % i, kwargs=kwargs)
        for i in range(num_threads)
    ]
    try:
        for thread in threads:
            thread.start()
    finally:
        for thread in threads:
            thread.join()

    # Retrieve exceptions
    exceptions = [exceptions.get(block=False) for i in range(num_threads)]
    return [exc for exc in exceptions if exc is not None]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import oscar.models.fields
from django.utils.module_loading import import_string
from django.conf import settings

models_AutoField = import_string(settings.DEFAULT_AUTO_FIELD)

class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Country',
            fields=[
                ('iso_3166_1_a2', models.CharField(primary_key=True, max_length=2, verbose_name='ISO 3166-1 alpha-2', serialize=False)),
                ('iso_3166_1_a3', models.CharField(max_length=3, verbose_name='ISO 3166-1 alpha-3', blank=True)),
                ('iso_3166_1_numeric', models.CharField(max_length=3, verbose_name='ISO 3166-1 numeric', blank=True)),
                ('printable_name', models.CharField(max_length=128, verbose_name='Country name')),
                ('name', models.CharField(max_length=128, verbose_name='Official name')),
                ('display_order', models.PositiveSmallIntegerField(default=0, verbose_name='Display order', db_index=True, help_text='Higher the number, higher the country in the list.')),
                ('is_shipping_country', models.BooleanField(default=False, db_index=True, verbose_name='Is shipping country')),
            ],
            options={
                'ordering': ('-display_order', 'printable_name'),
                'verbose_name_plural': 'Countries',
                'verbose_name': 'Country',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='UserAddress',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(verbose_name='Title', max_length=64, blank=True, choices=[('Mr', 'Mr'), ('Miss', 'Miss'), ('Mrs', 'Mrs'), ('Ms', 'Ms'), ('Dr', 'Dr')])),
                ('first_name', models.CharField(max_length=255, verbose_name='First name', blank=True)),
                ('last_name', models.CharField(max_length=255, verbose_name='Last name', blank=True)),
                ('line1', models.CharField(max_length=255, verbose_name='First line of address')),
                ('line2', models.CharField(max_length=255, verbose_name='Second line of address', blank=True)),
                ('line3', models.CharField(max_length=255, verbose_name='Third line of address', blank=True)),
                ('line4', models.CharField(max_length=255, verbose_name='City', blank=True)),
                ('state', models.CharField(max_length=255, verbose_name='State/County', blank=True)),
                ('postcode', oscar.models.fields.UppercaseCharField(max_length=64, verbose_name='Post/Zip-code', blank=True)),
                ('search_text', models.TextField(editable=False, verbose_name='Search text - used only for searching addresses')),
                ('phone_number', oscar.models.fields.PhoneNumberField(verbose_name='Phone number', help_text='In case we need to call you about your order', blank=True)),
                ('notes', models.TextField(verbose_name='Instructions', help_text='Tell us anything we should know when delivering your order.', blank=True)),
                ('is_default_for_shipping', models.BooleanField(default=False, verbose_name='Default shipping address?')),
                ('is_default_for_billing', models.BooleanField(default=False, verbose_name='Default billing address?')),
                ('num_orders', models.PositiveIntegerField(default=0, verbose_name='Number of Orders')),
                ('hash', models.CharField(max_length=255, editable=False, db_index=True, verbose_name='Address Hash')),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('country', models.ForeignKey(verbose_name='Country', to='address.Country', on_delete=models.CASCADE)),
                ('user', models.ForeignKey(verbose_name='User', related_name='addresses', to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE)),
            ],
            options={
                'ordering': ['-num_orders'],
                'verbose_name_plural': 'User addresses',
                'verbose_name': 'User address',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name='useraddress',
            unique_together=set([('user', 'hash')]),
        ),
    ]



from decimal import Decimal as D

from django import http
from django.contrib import messages
from django.core.exceptions import ImproperlyConfigured
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from oscar.core import prices
from oscar.core.loading import get_class, get_model

from . import exceptions

Repository = get_class("shipping.repository", "Repository")
SurchargeApplicator = get_class("checkout.applicator", "SurchargeApplicator")
OrderTotalCalculator = get_class("checkout.calculators", "OrderTotalCalculator")
CheckoutSessionData = get_class("checkout.utils", "CheckoutSessionData")
ShippingAddress = get_model("order", "ShippingAddress")
BillingAddress = get_model("order", "BillingAddress")
UserAddress = get_model("address", "UserAddress")


class CheckoutSessionMixin(object):
    """
    Mixin to provide common functionality shared between checkout views.

    All checkout views subclass this mixin. It ensures that all relevant
    checkout information is available in the template context.
    """

    # A pre-condition is a condition that MUST be met in order for a view
    # to be available. If it isn't then the customer should be redirected
    # to a view *earlier* in the chain.
    # pre_conditions is a list of method names that get executed before the
    # normal flow of the view. Each method should check some condition has been
    # met. If not, then an exception is raised that indicates the URL the
    # customer will be redirected to.

    pre_conditions = None

    # A *skip* condition is a condition that MUST NOT be met in order for a
    # view to be available. If the condition is met, this means the view MUST
    # be skipped and the customer should be redirected to a view *later* in
    # the chain.
    # Skip conditions work similar to pre-conditions, and get evaluated after
    # pre-conditions have been evaluated.
    skip_conditions = None

    def dispatch(self, request, *args, **kwargs):
        # Assign the checkout session manager so it's available in all checkout
        # views.
        self.checkout_session = CheckoutSessionData(request)

        # Check if this view should be skipped
        try:
            self.check_skip_conditions(request)
        except exceptions.PassedSkipCondition as e:
            return http.HttpResponseRedirect(e.url)

        # Enforce any pre-conditions for the view.
        try:
            self.check_pre_conditions(request)
        except exceptions.FailedPreCondition as e:
            for message in e.messages:
                messages.warning(request, message)
            return http.HttpResponseRedirect(e.url)

        return super().dispatch(request, *args, **kwargs)

    def check_pre_conditions(self, request):
        pre_conditions = self.get_pre_conditions(request)
        for method_name in pre_conditions:
            if not hasattr(self, method_name):
                raise ImproperlyConfigured(
                    "There is no method '%s' to call as a pre-condition" % (method_name)
                )
            getattr(self, method_name)(request)

    def get_pre_conditions(self, request):
        """
        Return the pre-condition method names to run for this view
        """
        if self.pre_conditions is None:
            return []
        return self.pre_conditions

    def check_skip_conditions(self, request):
        skip_conditions = self.get_skip_conditions(request)
        for method_name in skip_conditions:
            if not hasattr(self, method_name):
                raise ImproperlyConfigured(
                    "There is no method '%s' to call as a skip-condition"
                    % (method_name)
                )
            getattr(self, method_name)(request)

    def get_skip_conditions(self, request):
        """
        Return the skip-condition method names to run for this view
        """
        if self.skip_conditions is None:
            return []
        return self.skip_conditions

    # Re-usable pre-condition validators

    def check_basket_is_not_empty(self, request):
        if request.basket.is_empty:
            raise exceptions.FailedPreCondition(
                url=reverse("basket:summary"),
                message=_("You need to add some items to your basket to checkout"),
            )

    def check_basket_is_valid(self, request):
        """
        Check that the basket is permitted to be submitted as an order. That
        is, all the basket lines are available to buy - nothing has gone out of
        stock since it was added to the basket.
        """
        messages_list = []
        strategy = request.strategy
        for line in request.basket.all_lines():
            result = strategy.fetch_for_line(line)
            is_permitted, reason = result.availability.is_purchase_permitted(
                line.quantity
            )
            if not is_permitted:
                # Create a more meaningful message to show on the basket page
                msg = _(
                    "'%(title)s' is no longer available to buy (%(reason)s). "
                    "Please adjust your basket to continue"
                ) % {"title": line.product.get_title(), "reason": reason}
                messages_list.append(msg)
        if messages_list:
            raise exceptions.FailedPreCondition(
                url=reverse("basket:summary"), messages=messages_list
            )

    def check_user_email_is_captured(self, request):
        if (
            not request.user.is_authenticated
            and not self.checkout_session.get_guest_email()
        ):
            raise exceptions.FailedPreCondition(
                url=reverse("checkout:index"),
                message=_("Please either sign in or enter your email address"),
            )

    def check_shipping_data_is_captured(self, request):
        if not request.basket.is_shipping_required():
            # Even without shipping being required, we still need to check that
            # a shipping method code has been set.
            if not self.checkout_session.is_shipping_method_set(self.request.basket):
                raise exceptions.FailedPreCondition(
                    url=reverse("checkout:shipping-method"),
                )
            return

        # Basket requires shipping: check address and method are captured and
        # valid.
        self.check_a_valid_shipping_address_is_captured()
        self.check_a_valid_shipping_method_is_captured()

    def check_a_valid_shipping_address_is_captured(self):
        # Check that shipping address has been completed
        if not self.checkout_session.is_shipping_address_set():
            raise exceptions.FailedPreCondition(
                url=reverse("checkout:shipping-address"),
                message=_("Please choose a shipping address"),
            )

        # Check that the previously chosen shipping address is still valid
        shipping_address = self.get_shipping_address(basket=self.request.basket)
        if not shipping_address:
            raise exceptions.FailedPreCondition(
                url=reverse("checkout:shipping-address"),
                message=_(
                    "Your previously chosen shipping address is "
                    "no longer valid.  Please choose another one"
                ),
            )

    def check_a_valid_shipping_method_is_captured(self):
        # Check that shipping method has been set
        if not self.checkout_session.is_shipping_method_set(self.request.basket):
            raise exceptions.FailedPreCondition(
                url=reverse("checkout:shipping-method"),
                message=_("Please choose a shipping method"),
            )

        # Check that a *valid* shipping method has been set
        shipping_address = self.get_shipping_address(basket=self.request.basket)
        shipping_method = self.get_shipping_method(
            basket=self.request.basket, shipping_address=shipping_address
        )
        if not shipping_method:
            raise exceptions.FailedPreCondition(
                url=reverse("checkout:shipping-method"),
                message=_(
                    "Your previously chosen shipping method is "
                    "no longer valid.  Please choose another one"
                ),
            )

    def check_payment_data_is_captured(self, request):
        # We don't collect payment data by default so we don't have anything to
        # validate here. If your shop requires forms to be submitted on the
        # payment details page, then override this method to check that the
        # relevant data is available. Often just enforcing that the preview
        # view is only accessible from a POST request is sufficient.
        pass

    # Re-usable skip conditions

    def skip_unless_basket_requires_shipping(self, request):
        # Check to see that a shipping address is actually required.  It may
        # not be if the basket is purely downloads
        if not request.basket.is_shipping_required():
            raise exceptions.PassedSkipCondition(
                url=reverse("checkout:shipping-method")
            )

    def skip_unless_payment_is_required(self, request):
        # Check to see if payment is actually required for this order.
        shipping_address = self.get_shipping_address(request.basket)
        shipping_method = self.get_shipping_method(request.basket, shipping_address)
        if shipping_method:
            shipping_charge = shipping_method.calculate(request.basket)
        else:
            # It's unusual to get here as a shipping method should be set by
            # the time this skip-condition is called. In the absence of any
            # other evidence, we assume the shipping charge is zero.
            shipping_charge = prices.Price(
                currency=request.basket.currency, excl_tax=D("0.00"), tax=D("0.00")
            )

        surcharges = SurchargeApplicator(request).get_applicable_surcharges(
            basket=request.basket, shipping_charge=shipping_charge
        )
        total = self.get_order_totals(request.basket, shipping_charge, surcharges)
        if total.excl_tax == D("0.00"):
            raise exceptions.PassedSkipCondition(url=reverse("checkout:preview"))

    # Helpers

    def get_context_data(self, **kwargs):
        # Use the proposed submission as template context data.  Flatten the
        # order kwargs so they are easily available too.
        ctx = super().get_context_data()
        ctx.update(self.build_submission(**kwargs))
        ctx.update(kwargs)
        ctx.update(ctx["order_kwargs"])
        return ctx

    def build_submission(self, **kwargs):
        """
        Return a dict of data that contains everything required for an order
        submission.  This includes payment details (if any).

        This can be the right place to perform tax lookups and apply them to
        the basket.
        """
        # Pop the basket if there is one, because we pass it as a positional
        # argument to methods below
        basket = kwargs.pop("basket", self.request.basket)
        shipping_address = self.get_shipping_address(basket)
        shipping_method = self.get_shipping_method(basket, shipping_address)
        billing_address = self.get_billing_address(shipping_address)
        submission = {
            "user": self.request.user,
            "basket": basket,
            "shipping_address": shipping_address,
            "shipping_method": shipping_method,
            "billing_address": billing_address,
            "order_kwargs": {},
            "payment_kwargs": {},
        }

        if not shipping_method:
            total = shipping_charge = surcharges = None
        else:
            shipping_charge = shipping_method.calculate(basket)
            surcharges = SurchargeApplicator(
                self.request, submission
            ).get_applicable_surcharges(
                self.request.basket, shipping_charge=shipping_charge
            )
            total = self.get_order_totals(
                basket, shipping_charge=shipping_charge, surcharges=surcharges, **kwargs
            )

        submission["shipping_charge"] = shipping_charge
        submission["order_total"] = total
        submission["surcharges"] = surcharges

        # If there is a billing address, add it to the payment kwargs as calls
        # to payment gateways generally require the billing address. Note, that
        # it normally makes sense to pass the form instance that captures the
        # billing address information. That way, if payment fails, you can
        # render bound forms in the template to make re-submission easier.
        if billing_address:
            submission["payment_kwargs"]["billing_address"] = billing_address

        # Allow overrides to be passed in
        submission.update(kwargs)

        # Set guest email after overrides as we need to update the order_kwargs
        # entry.
        user = submission["user"]
        if (
            not user.is_authenticated
            and "guest_email" not in submission["order_kwargs"]
        ):
            email = self.checkout_session.get_guest_email()
            submission["order_kwargs"]["guest_email"] = email
        return submission

    def get_shipping_address(self, basket):
        """
        Return the (unsaved) shipping address for this checkout session.

        If the shipping address was entered manually, then we instantiate a
        ``ShippingAddress`` model with the appropriate form data (which is
        saved in the session).

        If the shipping address was selected from the user's address book,
        then we convert the ``UserAddress`` to a ``ShippingAddress``.

        The ``ShippingAddress`` instance is not saved as sometimes you need a
        shipping address instance before the order is placed.  For example, if
        you are submitting fraud information as part of a payment request.

        The ``OrderPlacementMixin.create_shipping_address`` method is
        responsible for saving a shipping address when an order is placed.
        """
        if not basket.is_shipping_required():
            return None

        addr_data = self.checkout_session.new_shipping_address_fields()
        if addr_data:
            # Load address data into a blank shipping address model
            return ShippingAddress(**addr_data)
        addr_id = self.checkout_session.shipping_user_address_id()
        if addr_id:
            try:
                address = UserAddress._default_manager.get(pk=addr_id)
            except UserAddress.DoesNotExist:
                # An address was selected but now it has disappeared.  This can
                # happen if the customer flushes their address book midway
                # through checkout.  No idea why they would do this but it can
                # happen.  Checkouts are highly vulnerable to race conditions
                # like this.
                return None
            else:
                # Copy user address data into a blank shipping address instance
                shipping_addr = ShippingAddress()
                address.populate_alternative_model(shipping_addr)
                return shipping_addr

    def get_shipping_method(self, basket, shipping_address=None, **kwargs):
        """
        Return the selected shipping method instance from this checkout session

        The shipping address is passed as we need to check that the method
        stored in the session is still valid for the shipping address.
        """
        code = self.checkout_session.shipping_method_code(basket)
        methods = Repository().get_shipping_methods(
            basket=basket,
            user=self.request.user,
            shipping_addr=shipping_address,
            request=self.request,
        )
        for method in methods:
            if method.code == code:
                return method

    def get_billing_address(self, shipping_address):
        """
        Return an unsaved instance of the billing address (if one exists)

        This method only returns a billing address if the session has been used
        to store billing address information. It's also possible to capture
        billing address information as part of the payment details forms, which
        never get stored in the session. In that circumstance, the billing
        address can be set directly in the build_submission dict.
        """
        if not self.checkout_session.is_billing_address_set():
            return None
        if self.checkout_session.is_billing_address_same_as_shipping():
            if shipping_address:
                address = BillingAddress()
                shipping_address.populate_alternative_model(address)
                return address

        addr_data = self.checkout_session.new_billing_address_fields()
        if addr_data:
            # A new billing address has been entered - load address data into a
            # blank billing address model.
            return BillingAddress(**addr_data)

        addr_id = self.checkout_session.billing_user_address_id()
        if addr_id:
            # An address from the user's address book has been selected as the
            # billing address - load it and convert it into a billing address
            # instance.
            try:
                user_address = UserAddress._default_manager.get(pk=addr_id)
            except UserAddress.DoesNotExist:
                # An address was selected but now it has disappeared.  This can
                # happen if the customer flushes their address book midway
                # through checkout.  No idea why they would do this but it can
                # happen.  Checkouts are highly vulnerable to race conditions
                # like this.
                return None
            else:
                # Copy user address data into a blank shipping address instance
                billing_address = BillingAddress()
                user_address.populate_alternative_model(billing_address)
                return billing_address

    def get_order_totals(self, basket, shipping_charge, surcharges=None, **kwargs):
        """
        Returns the total for the order with and without tax
        """
        return OrderTotalCalculator(self.request).calculate(
            basket, shipping_charge, surcharges, **kwargs
        )



from django.urls import path, re_path
from django.utils.translation import gettext_lazy as _

from oscar.core.application import OscarDashboardConfig
from oscar.core.loading import get_class


class CatalogueDashboardConfig(OscarDashboardConfig):
    label = "catalogue_dashboard"
    name = "oscar.apps.dashboard.catalogue"
    verbose_name = _("Catalogue")

    default_permissions = [
        "is_staff",
    ]

    def configure_permissions(self):
        DashboardPermission = get_class("dashboard.permissions", "DashboardPermission")

        self.permissions_map = {
            "catalogue-product": (
                DashboardPermission.get("product"),
                DashboardPermission.partner_dashboard_access,
            ),
            "catalogue-product-create": (
                DashboardPermission.get("product"),
                DashboardPermission.partner_dashboard_access,
            ),
            "catalogue-product-list": (
                DashboardPermission.get("product"),
                DashboardPermission.partner_dashboard_access,
            ),
            "catalogue-product-delete": (
                DashboardPermission.get("product"),
                DashboardPermission.partner_dashboard_access,
            ),
            "catalogue-product-lookup": (
                DashboardPermission.get("product"),
                DashboardPermission.partner_dashboard_access,
            ),
            "catalogue-product-create-child": DashboardPermission.get("product"),
            "stock-alert-list": DashboardPermission.get("stockalert"),
            "catalogue-category-list": DashboardPermission.get("category"),
            "catalogue-category-detail-list": DashboardPermission.get("category"),
            "catalogue-category-create": DashboardPermission.get("category"),
            "catalogue-category-create-child": DashboardPermission.get("category"),
            "catalogue-category-update": DashboardPermission.get("category"),
            "catalogue-category-delete": DashboardPermission.get("category"),
            "catalogue-class-create": DashboardPermission.get("product_class"),
            "catalogue-class-list": DashboardPermission.get("product_class"),
            "catalogue-class-update": DashboardPermission.get("product_class"),
            "catalogue-class-delete": DashboardPermission.get("product_class"),
            "catalogue-attribute-option-group-create": DashboardPermission.get(
                "attribute_option_group"
            ),
            "catalogue-attribute-option-group-list": DashboardPermission.get(
                "attribute_option_group"
            ),
            "catalogue-attribute-option-group-update": DashboardPermission.get(
                "attribute_option_group"
            ),
            "catalogue-attribute-option-group-delete": DashboardPermission.get(
                "attribute_option_group"
            ),
            "catalogue-option-list": DashboardPermission.get("option"),
            "catalogue-option-create": DashboardPermission.get("option"),
            "catalogue-option-update": DashboardPermission.get("option"),
            "catalogue-option-delete": DashboardPermission.get("option"),
        }

    # pylint: disable=attribute-defined-outside-init
    def ready(self):
        self.product_list_view = get_class(
            "dashboard.catalogue.views", "ProductListView"
        )
        self.product_lookup_view = get_class(
            "dashboard.catalogue.views", "ProductLookupView"
        )
        self.product_create_redirect_view = get_class(
            "dashboard.catalogue.views", "ProductCreateRedirectView"
        )
        self.product_createupdate_view = get_class(
            "dashboard.catalogue.views", "ProductCreateUpdateView"
        )
        self.product_delete_view = get_class(
            "dashboard.catalogue.views", "ProductDeleteView"
        )

        self.product_class_create_view = get_class(
            "dashboard.catalogue.views", "ProductClassCreateView"
        )
        self.product_class_update_view = get_class(
            "dashboard.catalogue.views", "ProductClassUpdateView"
        )
        self.product_class_list_view = get_class(
            "dashboard.catalogue.views", "ProductClassListView"
        )
        self.product_class_delete_view = get_class(
            "dashboard.catalogue.views", "ProductClassDeleteView"
        )

        self.category_list_view = get_class(
            "dashboard.catalogue.views", "CategoryListView"
        )
        self.category_detail_list_view = get_class(
            "dashboard.catalogue.views", "CategoryDetailListView"
        )
        self.category_create_view = get_class(
            "dashboard.catalogue.views", "CategoryCreateView"
        )
        self.category_update_view = get_class(
            "dashboard.catalogue.views", "CategoryUpdateView"
        )
        self.category_delete_view = get_class(
            "dashboard.catalogue.views", "CategoryDeleteView"
        )

        self.stock_alert_view = get_class(
            "dashboard.catalogue.views", "StockAlertListView"
        )

        self.attribute_option_group_create_view = get_class(
            "dashboard.catalogue.views", "AttributeOptionGroupCreateView"
        )
        self.attribute_option_group_list_view = get_class(
            "dashboard.catalogue.views", "AttributeOptionGroupListView"
        )
        self.attribute_option_group_update_view = get_class(
            "dashboard.catalogue.views", "AttributeOptionGroupUpdateView"
        )
        self.attribute_option_group_delete_view = get_class(
            "dashboard.catalogue.views", "AttributeOptionGroupDeleteView"
        )

        self.option_list_view = get_class("dashboard.catalogue.views", "OptionListView")
        self.option_create_view = get_class(
            "dashboard.catalogue.views", "OptionCreateView"
        )
        self.option_update_view = get_class(
            "dashboard.catalogue.views", "OptionUpdateView"
        )
        self.option_delete_view = get_class(
            "dashboard.catalogue.views", "OptionDeleteView"
        )
        self.configure_permissions()

    def get_urls(self):
        urls = [
            path(
                "products/<int:pk>/",
                self.product_createupdate_view.as_view(),
                name="catalogue-product",
            ),
            path(
                "products/create/",
                self.product_create_redirect_view.as_view(),
                name="catalogue-product-create",
            ),
            re_path(
                r"^products/create/(?P<product_class_slug>[\w-]+)/$",
                self.product_createupdate_view.as_view(),
                name="catalogue-product-create",
            ),
            path(
                "products/<int:parent_pk>/create-variant/",
                self.product_createupdate_view.as_view(),
                name="catalogue-product-create-child",
            ),
            path(
                "products/<int:pk>/delete/",
                self.product_delete_view.as_view(),
                name="catalogue-product-delete",
            ),
            path("", self.product_list_view.as_view(), name="catalogue-product-list"),
            path(
                "stock-alerts/",
                self.stock_alert_view.as_view(),
                name="stock-alert-list",
            ),
            path(
                "product-lookup/",
                self.product_lookup_view.as_view(),
                name="catalogue-product-lookup",
            ),
            path(
                "categories/",
                self.category_list_view.as_view(),
                name="catalogue-category-list",
            ),
            path(
                "categories/<int:pk>/",
                self.category_detail_list_view.as_view(),
                name="catalogue-category-detail-list",
            ),
            path(
                "categories/create/",
                self.category_create_view.as_view(),
                name="catalogue-category-create",
            ),
            path(
                "categories/create/<int:parent>/",
                self.category_create_view.as_view(),
                name="catalogue-category-create-child",
            ),
            path(
                "categories/<int:pk>/update/",
                self.category_update_view.as_view(),
                name="catalogue-category-update",
            ),
            path(
                "categories/<int:pk>/delete/",
                self.category_delete_view.as_view(),
                name="catalogue-category-delete",
            ),
            path(
                "product-type/create/",
                self.product_class_create_view.as_view(),
                name="catalogue-class-create",
            ),
            path(
                "product-types/",
                self.product_class_list_view.as_view(),
                name="catalogue-class-list",
            ),
            path(
                "product-type/<int:pk>/update/",
                self.product_class_update_view.as_view(),
                name="catalogue-class-update",
            ),
            path(
                "product-type/<int:pk>/delete/",
                self.product_class_delete_view.as_view(),
                name="catalogue-class-delete",
            ),
            path(
                "attribute-option-group/create/",
                self.attribute_option_group_create_view.as_view(),
                name="catalogue-attribute-option-group-create",
            ),
            path(
                "attribute-option-group/",
                self.attribute_option_group_list_view.as_view(),
                name="catalogue-attribute-option-group-list",
            ),
            # The RelatedFieldWidgetWrapper code does something funny with
            # placeholder urls, so it does need to match more than just a pk
            path(
                "attribute-option-group/<str:pk>/update/",
                self.attribute_option_group_update_view.as_view(),
                name="catalogue-attribute-option-group-update",
            ),
            # The RelatedFieldWidgetWrapper code does something funny with
            # placeholder urls, so it does need to match more than just a pk
            path(
                "attribute-option-group/<str:pk>/delete/",
                self.attribute_option_group_delete_view.as_view(),
                name="catalogue-attribute-option-group-delete",
            ),
            path(
                "option/", self.option_list_view.as_view(), name="catalogue-option-list"
            ),
            path(
                "option/create/",
                self.option_create_view.as_view(),
                name="catalogue-option-create",
            ),
            path(
                "option/<str:pk>/update/",
                self.option_update_view.as_view(),
                name="catalogue-option-update",
            ),
            path(
                "option/<str:pk>/delete/",
                self.option_delete_view.as_view(),
                name="catalogue-option-delete",
            ),
        ]
        return self.post_process_urls(urls)



# Use 'alpha', 'beta', 'rc' or 'final' as the 4th element to indicate release type.
VERSION = (4, 1, 0, "final")


def get_short_version():
    return "%s.%s" % (VERSION[0], VERSION[1])


def get_version():
    version = "%s.%s" % (VERSION[0], VERSION[1])
    # Append 3rd digit if > 0
    if VERSION[2]:
        version = "%s.%s" % (version, VERSION[2])

    if len(VERSION) > 3 and VERSION[3] != "final":
        mapping = {"alpha": "a", "beta": "b", "rc": "rc"}
        version = "%s%s" % (version, mapping[VERSION[3]])
        if len(VERSION) == 5:
            version = "%s%s" % (version, VERSION[4])

    return version


INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.sites",
    "django.contrib.flatpages",
    "oscar.config.Shop",
    "oscar.apps.analytics.apps.AnalyticsConfig",
    "oscar.apps.checkout.apps.CheckoutConfig",
    "oscar.apps.address.apps.AddressConfig",
    "oscar.apps.shipping.apps.ShippingConfig",
    "oscar.apps.catalogue.apps.CatalogueConfig",
    "oscar.apps.catalogue.reviews.apps.CatalogueReviewsConfig",
    "oscar.apps.communication.apps.CommunicationConfig",
    "oscar.apps.partner.apps.PartnerConfig",
    "oscar.apps.basket.apps.BasketConfig",
    "oscar.apps.payment.apps.PaymentConfig",
    "oscar.apps.offer.apps.OfferConfig",
    "oscar.apps.order.apps.OrderConfig",
    "oscar.apps.customer.apps.CustomerConfig",
    "oscar.apps.search.apps.SearchConfig",
    "oscar.apps.voucher.apps.VoucherConfig",
    "oscar.apps.wishlists.apps.WishlistsConfig",
    "oscar.apps.dashboard.apps.DashboardConfig",
    "oscar.apps.dashboard.reports.apps.ReportsDashboardConfig",
    "oscar.apps.dashboard.users.apps.UsersDashboardConfig",
    "oscar.apps.dashboard.orders.apps.OrdersDashboardConfig",
    "oscar.apps.dashboard.catalogue.apps.CatalogueDashboardConfig",
    "oscar.apps.dashboard.offers.apps.OffersDashboardConfig",
    "oscar.apps.dashboard.partners.apps.PartnersDashboardConfig",
    "oscar.apps.dashboard.pages.apps.PagesDashboardConfig",
    "oscar.apps.dashboard.ranges.apps.RangesDashboardConfig",
    "oscar.apps.dashboard.reviews.apps.ReviewsDashboardConfig",
    "oscar.apps.dashboard.vouchers.apps.VouchersDashboardConfig",
    "oscar.apps.dashboard.communications.apps.CommunicationsDashboardConfig",
    "oscar.apps.dashboard.shipping.apps.ShippingDashboardConfig",
    # 3rd-party apps that oscar depends on
    "widget_tweaks",
    "haystack",
    "treebeard",
    "django_tables2",
]


default_app_config = "oscar.config.Shop"



import datetime
import decimal
import logging
import re
import unicodedata

from babel.dates import format_timedelta as format_td
from django.conf import settings
from django.shortcuts import redirect, resolve_url
from django.template.defaultfilters import date as date_filter
from django.utils.http import url_has_allowed_host_and_scheme
from django.utils.module_loading import import_string
from django.utils.text import slugify as django_slugify
from django.utils.timezone import get_current_timezone, is_naive, make_aware
from django.utils.translation import get_language, to_locale

from oscar.core.decorators import deprecated

SLUGIFY_RE = re.compile(r"[^\w\s-]", re.UNICODE)


def cautious_slugify(value):
    """
    Convert a string to ASCII exactly as Django's slugify does, with the exception
    that any non-ASCII alphanumeric characters (that cannot be ASCIIfied under Unicode
    normalisation) are escaped into codes like 'u0421' instead of being deleted entirely.
    This ensures that the result of slugifying e.g. Cyrillic text will not be an empty
    string, and can thus be safely used as an identifier (albeit not a human-readable one).

    cautious_slugify was copied from Wagtail:
    <https://github.com/wagtail/wagtail/blob/8b420b9/wagtail/core/utils.py>

    Copyright (c) 2014-present Torchbox Ltd and individual contributors.
    Released under the BSD 3-clause "New" or "Revised" License
    <https://github.com/wagtail/wagtail/blob/8b420b9/LICENSE>

    Date: 2018-06-15
    """
    # Normalize the string to decomposed unicode form. This causes accented Latin
    # characters to be split into 'base character' + 'accent modifier'; the latter will
    # be stripped out by the regexp, resulting in an ASCII-clean character that doesn't
    # need to be escaped
    value = unicodedata.normalize("NFKD", value)

    # Strip out characters that aren't letterlike, underscores or hyphens,
    # using the same regexp that slugify uses. This ensures that non-ASCII non-letters
    # (e.g. accent modifiers, fancy punctuation) get stripped rather than escaped
    value = SLUGIFY_RE.sub("", value)

    # Encode as ASCII, escaping non-ASCII characters with backslashreplace, then convert
    # back to a unicode string (which is what slugify expects)
    value = value.encode("ascii", "backslashreplace").decode("ascii")

    # Pass to slugify to perform final conversion (whitespace stripping); this will
    # also strip out the backslashes from the 'backslashreplace' conversion
    return django_slugify(value)


def default_slugifier(value, allow_unicode=False):
    """
    Oscar's default slugifier function. When unicode is allowed
    it uses Django's slugify function, otherwise it uses cautious_slugify.
    """
    if allow_unicode:
        return django_slugify(value, allow_unicode=True)
    else:
        return cautious_slugify(value)


def slugify(value):
    """
    Slugify a string

    The OSCAR_SLUG_FUNCTION can be set with a dotted path to the slug
    function to use, defaults to 'oscar.core.utils.default_slugifier'.

    OSCAR_SLUG_MAP can be set of a dictionary of target:replacement pairs

    OSCAR_SLUG_BLACKLIST can be set to a iterable of words to remove after
    the slug is generated; though it will not reduce a slug to zero length.
    """
    value = str(value)

    # Re-map some strings to avoid important characters being stripped.  Eg
    # remap 'c++' to 'cpp' otherwise it will become 'c'.
    for k, v in settings.OSCAR_SLUG_MAP.items():
        value = value.replace(k, v)

    slugifier = import_string(settings.OSCAR_SLUG_FUNCTION)
    slug = slugifier(value, allow_unicode=settings.OSCAR_SLUG_ALLOW_UNICODE)

    # Remove stopwords from slug
    for word in settings.OSCAR_SLUG_BLACKLIST:
        slug = slug.replace(word + "-", "")
        slug = slug.replace("-" + word, "")

    return slug


def format_timedelta(td):
    """
    Takes an instance of timedelta and formats it as a readable translated string
    """
    return format_td(
        td, threshold=2, locale=to_locale(get_language() or settings.LANGUAGE_CODE)
    )


def format_datetime(dt, dt_format=None):
    """
    Takes an instance of datetime, converts it to the current timezone and
    formats it as a string. Use this instead of
    django.core.templatefilters.date, which expects localtime.

    :param format: Common will be settings.DATETIME_FORMAT or
                   settings.DATE_FORMAT, or the resp. shorthands
                   ('DATETIME_FORMAT', 'DATE_FORMAT')
    """
    if is_naive(dt):
        localtime = make_aware(dt, get_current_timezone())
        logging.warning("oscar.core.utils.format_datetime received native datetime")
    else:
        localtime = dt.astimezone(get_current_timezone())
    return date_filter(localtime, dt_format)


def datetime_combine(date, time):
    """Timezone aware version of `datetime.datetime.combine`"""
    return make_aware(datetime.datetime.combine(date, time), get_current_timezone())


def safe_referrer(request, default):
    """
    Takes the request and a default URL. Returns HTTP_REFERER if it's safe
    to use and set, and the default URL otherwise.

    The default URL can be a model with get_absolute_url defined, a urlname
    or a regular URL
    """
    referrer = request.META.get("HTTP_REFERER")
    if referrer and url_has_allowed_host_and_scheme(referrer, request.get_host()):
        return referrer
    if default:
        # Try to resolve. Can take a model instance, Django URL name or URL.
        return resolve_url(default)
    else:
        # Allow passing in '' and None as default
        return default


def redirect_to_referrer(request, default):
    """
    Takes request.META and a default URL to redirect to.

    Returns a HttpResponseRedirect to HTTP_REFERER if it exists and is a safe
    URL; to the default URL otherwise.
    """
    return redirect(safe_referrer(request, default))


def get_default_currency():
    """
    For use as the default value for currency fields.  Use of this function
    prevents Django's core migration engine from interpreting a change to
    OSCAR_DEFAULT_CURRENCY as something it needs to generate a migration for.
    """
    return settings.OSCAR_DEFAULT_CURRENCY


def round_half_up_two_dec(money):
    """
    Explicitly round a decimal to 2 places half up, as should be used for
    money.

    >>> exponent = decimal.Decimal('0.01')
    >>> should_not_be_one = decimal.Decimal('1.005')
    >>> should_not_be_one.quantize(exponent)
    Decimal('1.00')
    >>> round_half_up_two_dec(should_not_be_one)
    Decimal('1.01')
    """
    return money.quantize(decimal.Decimal("0.01"), decimal.ROUND_HALF_UP)


round_half_up = deprecated(round_half_up_two_dec)


def round_half_up_four_dec(money):
    """
    Explicitly round a decimal to 4 places half up, as should be used for
    money.

    >>> exponent = decimal.Decimal('0.0001')
    >>> should_not_be_one = decimal.Decimal('1.00005')
    >>> should_not_be_one.quantize(exponent)
    Decimal('1.00')
    >>> round_half_up_four_dec(should_not_be_one)
    Decimal('1.01')
    """
    return money.quantize(decimal.Decimal("0.0001"), decimal.ROUND_HALF_UP)


def is_ajax(request):
    """
    Django 3.1 deprecated the  HttpRequest.is_ajax method as it relies on
    a jQuery-specific way of signifying AJAX calls.

    https://docs.djangoproject.com/en/3.1/releases/3.1/#id2

    For Oscar projects that's usually not a concern though.
    """
    return request.META.get("HTTP_X_REQUESTED_WITH") == "XMLHttpRequest"



import warnings

from django import template
from django.template.base import TextNode

from oscar.utils.deprecation import RemovedInOscar32Warning

register = template.Library()


# pylint: disable=unused-argument
@register.tag
def annotate_form_field(parser, token):
    """
    Used to set an attribute on a form field with the widget type. This is now
    done by Django itself.
    """
    warnings.warn(
        "The annotate_form_field template tag is deprecated and will be removed in the next version of django-oscar",
        RemovedInOscar32Warning,
        stacklevel=2,
    )
    return TextNode("")



from django.contrib.auth.tokens import default_token_generator
from django.urls import reverse
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_encode

from oscar.core.loading import get_class

Dispatcher = get_class("communication.utils", "Dispatcher")
Selector = get_class("partner.strategy", "Selector")


class CustomerDispatcher:
    """
    Dispatcher to send concrete customer related emails.
    """

    # Event codes
    REGISTRATION_EVENT_CODE = "REGISTRATION"
    PASSWORD_RESET_EVENT_CODE = "PASSWORD_RESET"
    PASSWORD_CHANGED_EVENT_CODE = "PASSWORD_CHANGED"
    EMAIL_CHANGED_EVENT_CODE = "EMAIL_CHANGED"

    def __init__(self, logger=None, mail_connection=None):
        self.dispatcher = Dispatcher(logger=logger, mail_connection=mail_connection)

    def send_registration_email_for_user(self, user, extra_context):
        messages = self.dispatcher.get_messages(
            self.REGISTRATION_EVENT_CODE, extra_context
        )
        self.dispatcher.dispatch_user_messages(user, messages)

    def send_password_reset_email_for_user(self, user, extra_context):
        messages = self.dispatcher.get_messages(
            self.PASSWORD_RESET_EVENT_CODE, extra_context
        )
        self.dispatcher.dispatch_user_messages(user, messages)

    def send_password_changed_email_for_user(self, user, extra_context):
        messages = self.dispatcher.get_messages(
            self.PASSWORD_CHANGED_EVENT_CODE, extra_context
        )
        self.dispatcher.dispatch_user_messages(user, messages)

    def send_email_changed_email_for_user(self, user, extra_context):
        messages = self.dispatcher.get_messages(
            self.EMAIL_CHANGED_EVENT_CODE, extra_context
        )
        self.dispatcher.dispatch_user_messages(user, messages)


def get_password_reset_url(user, token_generator=default_token_generator):
    """
    Generate a password-reset URL for a given user
    """
    kwargs = {
        "token": token_generator.make_token(user),
        "uidb64": urlsafe_base64_encode(force_bytes(user.id)),
    }
    return reverse("password-reset-confirm", kwargs=kwargs)


def normalise_email(email):
    """
    The local part of an email address is case-sensitive, the domain part
    isn't.  This function lowercases the host and should be used in all email
    handling.
    """
    clean_email = email.strip()
    if "@" in clean_email:
        local, host = clean_email.rsplit("@", 1)
        return local + "@" + host.lower()
    return clean_email



# -*- coding: utf-8 -*-
import sys

from django.core.management.base import BaseCommand, CommandError

from oscar.core.loading import get_model

Country = get_model("address", "Country")


class Command(BaseCommand):
    help = "Populates the list of countries with data from pycountry."

    def add_arguments(self, parser):
        parser.add_argument(
            "--no-shipping",
            action="store_false",
            dest="is_shipping",
            default=True,
            help="Don't mark countries for shipping",
        )
        parser.add_argument(
            "--initial-only",
            action="store_true",
            dest="is_initial_only",
            default=False,
            help="Exit quietly without doing anything if countries were already populated.",
        )

    def handle(self, *args, **options):
        try:
            import pycountry
        except ImportError:
            raise CommandError(
                "You are missing the pycountry library. Install it with "
                "'pip install pycountry'"
            )

        if Country.objects.exists():
            if options.get("is_initial_only", False):
                # exit quietly, as the initial load already seems to have happened.
                self.stdout.write("Countries already populated; nothing to be done.")
                sys.exit(0)
            else:
                raise CommandError(
                    "You already have countries in your database. This command"
                    " currently does not support updating existing countries."
                )

        countries = [
            Country(
                iso_3166_1_a2=country.alpha_2,
                iso_3166_1_a3=country.alpha_3,
                iso_3166_1_numeric=country.numeric,
                printable_name=country.name,
                name=getattr(country, "official_name", ""),
                is_shipping_country=options["is_shipping"],
            )
            for country in pycountry.countries
        ]

        Country.objects.bulk_create(countries)
        self.stdout.write("Successfully added %s countries." % len(countries))



from django import forms
from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext_lazy

from oscar.core.loading import get_model
from oscar.core.validators import URLDoesNotExistValidator

FlatPage = get_model("flatpages", "FlatPage")


class PageSearchForm(forms.Form):
    """
    Search form to filter pages by *title.
    """

    title = forms.CharField(required=False, label=pgettext_lazy("Page title", "Title"))


class PageUpdateForm(forms.ModelForm):
    """
    Update form to create/update flatpages. It provides a *title*, *url*,
    and *content* field. The specified URL will be validated and check if
    the same URL already exists in the system.
    """

    url = forms.RegexField(
        label=_("URL"),
        max_length=100,
        regex=r"^[-\w/\.~]+$",
        required=False,
        help_text=_("Example: '/about/contact/'."),
        error_messages={
            "invalid": _(
                "This value must contain only letters, numbers, dots, "
                "underscores, dashes, slashes or tildes."
            ),
        },
    )

    def clean_url(self):
        """
        Validate the input for field *url* checking if the specified
        URL already exists. If it is an actual update and the URL has
        not been changed, validation will be skipped.

        Returns cleaned URL or raises an exception.
        """
        url = self.cleaned_data["url"]
        if "url" in self.changed_data:
            if not url.endswith("/"):
                url += "/"
            URLDoesNotExistValidator()(url)
        return url

    class Meta:
        model = FlatPage
        fields = ("title", "url", "content")



# Generated by Django 3.0.3 on 2020-03-03 09:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('basket', '0008_auto_20181115_1953'),
    ]

    operations = [
        migrations.AddField(
            model_name='line',
            name='date_updated',
            field=models.DateTimeField(auto_now=True, db_index=True, verbose_name='Date Updated'),
        ),
    ]



from oscar.apps.customer import abstract_models
from oscar.core.loading import is_model_registered

__all__ = []


if not is_model_registered("customer", "ProductAlert"):

    class ProductAlert(abstract_models.AbstractProductAlert):
        pass

    __all__.append("ProductAlert")



# Generated by Django 3.0.8 on 2020-08-01 07:17

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('offer', '0008_auto_20181115_1953'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='range',
            options={'ordering': ['name'], 'verbose_name': 'Range', 'verbose_name_plural': 'Ranges'},
        ),
    ]



from django.db.models import F

from oscar.apps.basket.signals import voucher_addition, voucher_removal


# pylint: disable=unused-argument
def track_voucher_addition(basket, voucher, **kwargs):
    voucher.num_basket_additions += 1
    voucher.__class__._default_manager.filter(pk=voucher.pk).update(
        num_basket_additions=F("num_basket_additions") + 1,
    )


# pylint: disable=unused-argument
def track_voucher_removal(basket, voucher, **kwargs):
    voucher.num_basket_additions -= 1
    voucher.__class__._default_manager.filter(pk=voucher.pk).update(
        num_basket_additions=F("num_basket_additions") - 1,
    )


voucher_addition.connect(track_voucher_addition)
voucher_removal.connect(track_voucher_removal)
